from importlib import resources
import pickle
import os
import requests
import io

import pandas as pnd
import gempipe
import cobra

from .repeating import check_gpr, check_author, check_rstring_arrow, add_reaction, get_deprecated_kos



def get_db(logger):
    
    
    logger.info("Downloading the database...")
    sheet_id = "1dXJBIFjCghrdvQtxEOYlVNWAQU4mK-nqLWyDQeUZqek"
    url = f"https://docs.google.com/spreadsheets/d/{sheet_id}/export?format=xlsx"
    response = requests.get(url)  # download the requested file
    if response.status_code == 200:
        excel_data = io.BytesIO(response.content)   # load into memory
        exceldb = pnd.ExcelFile(excel_data)
    else:
        logger.error(f"Error during download. Please contact the developer.")
        return 1
    
    
    # check table presence
    sheet_names = exceldb.sheet_names
    for i in ['T', 'R', 'M', 'authors']: 
        if i not in sheet_names:
            logger.error(f"Sheet '{i}' is missing!")
            return 1
        
        
    # load the tables
    db = {}
    db['T'] = exceldb.parse('T')
    db['R'] = exceldb.parse('R')
    db['M'] = exceldb.parse('M')
    db['authors'] = exceldb.parse('authors')
    
    
    # check table headers
    headers = {}
    headers['T'] = ['rid', 'rstring', 'kr', 'gpr_manual', 'name', 'author', 'notes']
    headers['R'] = ['rid', 'rstring', 'kr', 'gpr_manual', 'name', 'author', 'notes']
    headers['M'] = ['pure_mid', 'formula', 'charge', 'kc', 'name', 'inchikey', 'author', 'notes']
    headers['authors'] = ['username', 'first_name', 'last_name', 'role', 'mail']
    for i in db.keys(): 
        if set(db[i].columns) != set(headers[i]):
            logger.error(f"Sheet '{i}' is missing the columns {set(headers[i]) - set(db[i].columns)}.")
            return 1
        
    return db
    

    
def introduce_metabolites(logger, db, model, idcollection_dict, goodbefore, onlyauthor):
    goodbefore_reached = False

    
    logger.debug("Checking duplicated metabolite IDs...")
    if len(set(db['M']['pure_mid'].to_list())) != len(db['M']): 
        pure_mids = db['M']['pure_mid'].to_list()
        duplicates = list(set([item for item in pure_mids if pure_mids.count(item) > 1]))
        logger.error(f"Sheet 'M' has duplicated metabolites: {duplicates}.")
        return 1
   
        
    # parse M:
    logger.debug("Parsing metabolites...")
    db['M'] = db['M'].set_index('pure_mid', drop=True, verify_integrity=True)
    kc_ids_modeled = set()   # account for kc codes modeled
    for pure_mid, row in db['M'].iterrows():
        
        
        # skip empty lines!
        if type(pure_mid) != str: continue
        if pure_mid.strip() == '': continue
        if pure_mid == goodbefore:
            goodbefore_reached = True
        
        
        # parse author
        response = check_author(logger, pure_mid, row, db, 'M')
        if type(response) == int: return 1
        else: authors = response
        
        
        # manage goodbefore/onlyauthor
        if goodbefore != None and goodbefore_reached:
            if onlyauthor != None and onlyauthor not in authors:
                authors_string = '; '.join(authors)
                logger.info(f"Skipping metabolite '{pure_mid}' (authors '{authors_string}') as requested with --goodbefore[0] '{goodbefore}' and --onlyauthor '{onlyauthor}'.")
                continue
            if onlyauthor == None:
                logger.info(f"Skipping metabolite '{pure_mid}' as requested with --goodbefore[0] '{goodbefore}'.")
                continue
        
        
        # parse formula:
        if pnd.isna(row['formula']):
            logger.error(f"Metabolite '{pure_mid}' has missing formula: '{row['formula']}'.")
            return 1
  
        
        # parse charge: 
        if pnd.isna(row['charge']): 
            logger.error(f"Metabolite '{pure_mid}' has missing charge: '{row['charge']}'.")
            return 1
        
        
        # check if 'kc' codes are real:
        if pnd.isna(row['kc']): 
            logger.error(f"Metabolite '{pure_mid}' has missing KEGG annotation (kc): '{row['kc']}'.")
            return 1
        kc_ids = row['kc'].split(';')
        kc_ids = [i.strip() for i in kc_ids]
        for kc_id in kc_ids:
            if kc_id == 'CXXXXX':  # not in KEGG; could be knowledge gap (e.g. methyl group acceptor in R10404)
                logger.debug(f"Metabolite '{pure_mid}' is not in KEGG ('{kc_id}')!")
                continue  
            if kc_id not in idcollection_dict['kc']:
                logger.error(f"Metabolite '{pure_mid}' has invalid KEGG annotation (kc): '{kc_id}'.")
                return 1
            if kc_id in kc_ids_modeled:
                logger.error(f"KEGG annotation (kc) '{kc_id}' used in metabolite '{pure_mid}' is duplicated.")
                return 1
            if kc_id != 'CXXXXX':
                kc_ids_modeled.add(kc_id)
            
            
        # check the existance of the inchikey
        if pnd.isna(row['inchikey']): 
            logger.error(f"Metabolite '{pure_mid}' has missing inchikey: '{row['inchikey']}'.")
            return 1
        # check inchikey format:
        if len(row['inchikey']) != 27 or row['inchikey'][14] != '-' or row['inchikey'][25] != '-':
            logger.error(f"Metabolite '{pure_mid}' has badly formatted inchikey: '{row['inchikey']}'.")
            return 1
        
        
        # add metabolite to model
        m = cobra.Metabolite(f'{pure_mid}_c')
        model.add_metabolites([m])
        m = model.metabolites.get_by_id(f'{pure_mid}_c')
        m.name = row['name'].strip()
        m.formula = row['formula']
        m.charge = row['charge']
        m.compartment='c'
        # add kc annotations to model
        m.annotation['kegg.compound'] = kc_ids
        
    
    if goodbefore != None and goodbefore_reached == False:
        logger.info(f"Metabolite '{goodbefore}' never reached. Are you sure about your --goodbefore?")
                    
                    
    return model
    
    
    
def introduce_reactions(logger, db, model, idcollection_dict, goodbefore, onlyauthor): 
    goodbefore_reached = False
                    
    
    
    logger.debug("Checking duplicated reaction IDs...")
    if len(set(db['R']['rid'].to_list())) != len(db['R']): 
        pure_mids = db['R']['rid'].to_list()
        duplicates = list(set([item for item in pure_mids if pure_mids.count(item) > 1]))
        logger.error(f"Sheet 'R' has duplicated reactions: {duplicates}.")
        return 1
    
        
    # parse R:
    logger.debug("Parsing reactions...")
    db['R'] = db['R'].set_index('rid', drop=True, verify_integrity=True)
    for rid, row in db['R'].iterrows():
        
        
        # skip empty lines!
        if type(rid) != str: continue
        if rid.strip() == '': continue
        if rid == goodbefore:
            goodbefore_reached = True
        
        
        # parse author
        response = check_author(logger, rid, row, db, 'R')
        if type(response) == int: return 1
        else: authors = response
                    
                    
        # manage goodbefore/onlyauthor
        if goodbefore != None and goodbefore_reached:
            if onlyauthor != None and onlyauthor not in authors:
                authors_string = '; '.join(authors)
                logger.info(f"Skipping reaction '{rid}' (authors '{authors_string}') as requested with --goodbefore[1] '{goodbefore}' and --onlyauthor '{onlyauthor}'.")
                continue
            if onlyauthor == None:
                logger.info(f"Skipping reaction '{rid}' as requested with --goodbefore[1] '{goodbefore}'.")
                continue
        
        
        # parse reaction string
        response = check_rstring_arrow(logger, row, 'R')
        if response == 1: return 1
        

        # check if 'kr' codes are real:
        if pnd.isna(row['kr']): 
            logger.error(f"Reaction '{rid}' has missing KEGG annotation (kr): '{row['kr']}'.")
            return 1
        kr_ids = row['kr'].split(';')
        kr_ids = [i.strip() for i in kr_ids]
        for kr_id in kr_ids:
            if kr_id == 'RXXXXX':  # not in KEGG; could be knowledge gap 
                logger.debug(f"Reaction '{rid}' is not in KEGG ('{kr_id}')!")
                continue  
            if kr_id not in idcollection_dict['kr']:
                logger.error(f"Reaction '{rid}' has invalid KEGG annotation (kr): '{kr_id}'.")
                return 1
        
            
        # check GPR:
        response = check_gpr(logger, rid, row, kr_ids, idcollection_dict, 'R')
        if response == 1: return 1
        
        
        # add reaction to model
        response = add_reaction(logger, model, rid, row, kr_ids, 'R')
        if response == 1: return 1
               
    
    if goodbefore != None and goodbefore_reached == False:
        logger.info(f"Reaction '{goodbefore}' never reached. Are you sure about your --goodbefore?")
    
    
    return model
      
    
    
def introduce_transporters(logger, db, model, idcollection_dict, goodbefore, onlyauthor): 
    goodbefore_reached = False
                    
    
    
    def clone_to_external(model, mid_c, mid_e):
    
        m = cobra.Metabolite(f'{mid_e}')
        model.add_metabolites([m])
        
        m_c = model.metabolites.get_by_id(f'{mid_c}')
        m_e = model.metabolites.get_by_id(f'{mid_e}')
        m_e.compartment='e'
        
        m_e.name = m_c.name
        m_e.formula = m_c.formula
        m_e.charge = m_c.charge
        m_e.annotation = m_c.annotation
            
    
    def add_exchange_reaction(model, mid_e):
        
        r = cobra.Reaction(f'EX_{mid_e}')
        model.add_reactions([r])
        r = model.reactions.get_by_id(f'EX_{mid_e}')
        r.name = f"Exchange for {model.metabolites.get_by_id(mid_e).name}"
        r.build_reaction_from_string(f'{mid_e} --> ')
        if mid_e in [
            # basics:
            'glc__D_e', 'nh4_e', 'pi_e', 'so4_e', 'h2o_e', 'h_e', 'o2_e', 'co2_e',
            # metals: 
            'cu2_e', 'mobd_e', 'fe2_e', 'cobalt2_e',
        ]:
            r.bounds = (-1000, 1000)
        else:
            r.bounds = (0, 1000)
    
    
    
    # get all already inserted metabolites
    mids_parsed = [m.id for m in model.metabolites]
    rids_parsed = [r.id for r in model.reactions]
    
    
    # protons may not have an explicit transporter
    clone_to_external(model, 'h_c', 'h_e')
    mids_parsed.append('h_e')
    add_exchange_reaction(model, 'h_e')
    rids_parsed.append(f'EX_h_e')
    
    
    # parse T:
    logger.debug("Parsing transporters...")
    db['T'] = db['T'].set_index('rid', drop=True, verify_integrity=True)
    for rid, row in db['T'].iterrows():
        
        
        # skip empty lines!
        if type(rid) != str: continue
        if rid.strip() == '': continue
        if rid == goodbefore:
            goodbefore_reached = True
        
        
        # parse author
        response = check_author(logger, rid, row, db, 'T')
        if type(response) == int: return 1
        else: authors = response
                    
                    
        # manage goodbefore/onlyauthor
        if goodbefore != None and goodbefore_reached:
            if onlyauthor != None and onlyauthor not in authors:
                authors_string = '; '.join(authors)
                logger.info(f"Skipping transport '{rid}' (authors '{authors_string}') as requested with --goodbefore[2] '{goodbefore}' and --onlyauthor '{onlyauthor}'.")
                continue
            if onlyauthor == None:
                logger.info(f"Skipping transport '{rid}' as requested with --goodbefore[2] '{goodbefore}'.")
                continue
        
        
        # parse reaction string
        response = check_rstring_arrow(logger, row, 'T')
        if response == 1: return 1
    
    
        # check if 'kr' codes are real:
        if pnd.isna(row['kr']): 
            logger.error(f"Reaction '{rid}' has missing KEGG annotation (kr): '{row['kr']}'.")
            return 1
        if row['kr'] != '-':
            kr_ids = row['kr'].split(';')
            kr_ids = [i.strip() for i in kr_ids]
            for kr_id in kr_ids:
                if kr_id not in idcollection_dict['kr']:
                    logger.error(f"Reaction '{rid}' has invalid KEGG annotation (kr): '{kr_id}'.")
                    return 1
        else: kr_ids = []

            
        # check GPR:
        response = check_gpr(logger, rid, row, kr_ids, idcollection_dict, 'T')
        if response == 1: return 1
        
        
        # get involved metabolites:
        involved_mids = row['rstring'].split(' ')
        involved_mids = [i for i in involved_mids if i not in ['-->', '<=>']]
        
        
        # the external metabolite must be already modeled as cytosolic
        for mid in involved_mids: 
            if mid.endswith('_e'):
                mid_e = mid
                mid_c = mid.rsplit('_', 1)[0] + '_c'
                if mid_c not in mids_parsed:
                    logger.error(f"{rid}: the metabolite '{mid_c}', counterpart of '{mid_e}', was not previously modeled.")
                    return 1
                
                # add external metabolite to model
                if mid_e not in mids_parsed:
                    clone_to_external(model, mid_c, mid_e)
                    mids_parsed.append(mid_e)
                    
                # add exchange reaction to model
                if f'EX_{mid_e}' not in rids_parsed:
                    add_exchange_reaction(model, mid_e)
                    rids_parsed.append(f'EX_{mid_e}')
                    
                    
        # add reaction to model
        response = add_reaction(logger, model, rid, row, kr_ids, 'T')
        if response == 1: return 1
        
        
    if goodbefore != None and goodbefore_reached == False:
        logger.info(f"Transport '{goodbefore}' never reached. Are you sure about your --goodbefore?")
        
    
    return model



def introduce_sinks_demands(logger, model): 
    
    sinks = ['apoACP', 'apocarb', 'thioca', 'THI5p_b', 'cyE']
    demands = ['scp', 'amob', 'dialurate', 'THI5p_a' ]
    
    
    for puremid in sinks: 
        r = cobra.Reaction(f'sn_{puremid}_c')
        model.add_reactions([r])
        r = model.reactions.get_by_id(f'sn_{puremid}_c')
        r.name = f"Sink for {model.metabolites.get_by_id(f'{puremid}_c').name}"
        r.build_reaction_from_string(f'{puremid}_c <=> ')
        r.bounds = (-1000, 1000)
    
    
    for puremid in demands: 
        r = cobra.Reaction(f'dm_{puremid}_c')
        model.add_reactions([r])
        r = model.reactions.get_by_id(f'dm_{puremid}_c')
        r.name = f"Demand for {model.metabolites.get_by_id(f'{puremid}_c').name}"
        r.build_reaction_from_string(f'{puremid}_c --> ')
        r.bounds = (0, 1000)
    
    
    return model



def introduce_biomass(logger, db, model): 
    
    
    # Note: universal and conditional precursors have been defined in Xavier2017 .
    # Here presented in the same order of Xavier2017 .
    # Xavier2017: 10.1016/j.ymben.2016.12.002.
    biomass_dict = {
        ##### UNIVERSAL #####
        'ribo_nucleotides': [
            'atp_c', 'ctp_c', 'gtp_c', 'utp_c'
        ],
        'deoxyribo_nucleotides': [
            'datp_c', 'dctp_c', 'dgtp_c', 'dttp_c'
        ],
        'amino_acids': [
            'ala__L_c', 'arg__L_c', 'asn__L_c', 'asp__L_c', 'cys__L_c', 
            'gln__L_c', 'glu__L_c', 'gly_c',    'his__L_c', 'ile__L_c', 
            'leu__L_c', 'lys__L_c', 'met__L_c', 'ser__L_c', 'pro__L_c', 
            'thr__L_c', 'trp__L_c', 'tyr__L_c', 'val__L_c', 'phe__L_c'
        ],
        'cofactors_uni': [
            'nad_c',    # B3: Nicotinamide -adenine dinucleotide phosphate
            'nadp_c',   # B3: Nicotinamide -adenine dinucleotide phosphate
            'coa_c',    # B5: Coenzyme A  (dependant on 'pnto__R_c')
            'fad_c',    # B2: Flavin adenine dinucleotide
            'fmn_c',    # B2: Flavin mononucleotide
            'ribflv_c', # B2: ribovlavin. Non-active form acording to Xavier2017 but anyway included.
            #'f4200_c', # B2: included by Xavier2017 in 'universal' but the description seems conditional.
            'thf_c',    # B9: tetrahydrofolate 
            '10fthf_c', # B9: 10-Formyltetrahydrofolate
            '5mthf_c',  # B9: 5-Methyltetrahydrofolate
            'thmpp_c',  # B1: Thiamine diphosphate
            'pydx5p_c', # B6: pyridoxal 5-phosphate
            'amet_c',   # SAM: S-adenosyl-methionine
        ],
        ##### CONDITIONAL #####
        'cofactors_con': [
            #'f4200_c', # coenzyme f420 (electron transfer in methanogens, actinobacteria , and others)
            'ptrc_c',   # Putrescine
            'spmd_c',   # Sperimidine
            'pheme_c',  # protoheme
            'mql8_c',   # menaquinol / manaquinone (mqn8_c)
            'q8h2_c',   # ubiquinol / ubiquinone (q8_c)
            # Methionaquinone
            'btn_c',    # B7: biotin
            'ACP_c',    # Acyl-carrier protein
            'adocbl_c', # B12: Adenosylcobalamin
            # Lipoate
            'uacgam_c'  # uridine diphosphate N-Acetylglucosamine (UDP-GlcNAc)
        ],
        ##### ADDED ##### (conditionals not included or lumped in Xavier2017)
        'cofactors_add': [
            'hemeO_c',  # heme-O
            'sheme_c',  # siroheme
            'moco_c',   # molybdenum cofactor
            'phllqol_c',# phylloquinol / phylloquinone (phllqne_c)
            'gthrd_c',  # glutathione (reduced)
        ],
        'lipids': [
            'pe120_c', # phosphatidyl-ethanolamine (12:0;12:0)
            'pg120_c', # phosphatidyl-glycerol (12:0;12:0)
            'clpn120_c', # cardiolipin (12:0;12:0;12:0;12:0)
            # 1-lysyl phosphatidylglycerol (plantarum)
        ],
        'membrane_wall': [
            'peptidoSTA_c', # peptidoglycan (dependant on 'udcpdp_c')
            'WTAgg40r_20n_20a_P_c', # teichoic acids
            'WTArg40r_20g_20a_P_c', # teichoic acids
            'WTAg40g_20g_20a_P_c', # teichoic acids
            'LTAgg40g_20n_20a_c', # lipoteichoic acids
            'LTAga40g_20t_20a_c', # lipoteichoic acids            
            # capsular polysaccharides
            # kdo_lipid_A
        ],
        'energy_stock': [
            # glycogen
            # starch
            # PHA / PHB
        ]
    }
    
    
    rstring =           f'0.01 {" + 0.01 ".join(biomass_dict["ribo_nucleotides"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["deoxyribo_nucleotides"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["amino_acids"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["cofactors_uni"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["cofactors_con"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["cofactors_add"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["lipids"])}'
    rstring = rstring + f' + 0.01 {" + 0.01 ".join(biomass_dict["membrane_wall"])}'
    rstring = rstring + f' + 0.01 atp_c + 0.01 h2o_c --> 0.01 adp_c + 0.01 h_c + 0.01 pi_c'
    
    
    r = cobra.Reaction('Biomass')
    model.add_reactions([r])
    r = model.reactions.get_by_id('Biomass')
    r.name = 'Biomass assembly reaction'
    r.build_reaction_from_string(rstring)
    
    
    # set as objective:
    model.objective = 'Biomass'
    
    
    return model



def translate_annotate_genes(logger, model, idcollection_dict):
    

       
    ko_to_name = idcollection_dict['ko_to_name']
    ko_to_symbols = idcollection_dict['ko_to_symbols']
    ko_to_ecs = idcollection_dict['ko_to_ecs']
    ko_to_cogs = idcollection_dict['ko_to_cogs']
    ko_to_gos = idcollection_dict['ko_to_gos']
    
    
    # translation dicts: assign to each KO a symbol that is unique in the universe model.
    ko_to_sym = {}
    sym_to_ko = {}
    cnt = 0
    for g in model.genes:
        if g.id in ['orphan', 'spontaneous']: 
            continue
        ko = g.id
        cnt += 1
        
        if ko in get_deprecated_kos():
            # if the ko is deprecated, it was not included in 'ko_to_symbols'
            ko_to_sym[ko] = ko
            sym_to_ko[ko] = ko
            continue
            
        for symbol in ko_to_symbols[ko]:  # iterate the available symbols for this KO
            if symbol not in sym_to_ko.keys():   # take the first available (not yet used)
                ko_to_sym[ko] = symbol
                sym_to_ko[symbol] = ko
                break
        
        if cnt != len(ko_to_sym):  # no symbol was assigned (symbol was already taken by another KO)
            cnt_dups = 2
            symbol = list(ko_to_symbols[ko])[0] + f'_{cnt_dups}'   # generate a new symbol
            while cnt != len(ko_to_sym):   # until a symbol is assigned
                if symbol not in sym_to_ko.keys():   # if the new symbol fits
                    ko_to_sym[ko] = symbol
                    sym_to_ko[symbol] = ko
                cnt_dups += 1
                symbol = list(ko_to_symbols[ko])[0] + f'_{cnt_dups}'   # retry with the next one
                

                
    
    # insert annotations
    for g in model.genes:
        if g.id in ['orphan', 'spontaneous']: 
            continue
        ko = g.id
        g.annotation['ko'] = ko
        
        if ko not in get_deprecated_kos():
            # deprecated kos are missing from these dicts
            g.name = ko_to_name[ko]
            g.annotation['symbols'] = list(ko_to_symbols[ko])
            g.annotation['ec'] = list(ko_to_ecs[ko])
            g.annotation['cog'] = list(ko_to_cogs[ko])
            g.annotation['go'] = list(ko_to_gos[ko])
        
    
        
    # finally apply translations of IDs
    translation_dict = ko_to_sym
    translation_dict['orphan'] = 'orphan'
    translation_dict['spontaneous'] = 'spontaneous'
    cobra.manipulation.rename_genes(model, translation_dict)
    
    
    return model
    


def set_up_groups(logger, model, idcollection_dict):
    

       
    kr_to_maps = idcollection_dict['kr_to_maps']
    map_to_name = idcollection_dict['map_to_name']
    kr_to_mds = idcollection_dict['kr_to_mds']
    md_to_name = idcollection_dict['md_to_name']
    
    
    # define groups of available contents
    groups = {}   # mixing maps and mds
    for r in model.reactions:
        
        if 'kegg.reaction' not in r.annotation.keys():
            continue   # Biomass, exchanges, demands, sinks, transporters
        kr_ids = r.annotation['kegg.reaction']
            
        for kr_id in kr_ids:
            if kr_id == 'RXXXXX':
                continue
            
            # insert maps
            for map_id in kr_to_maps[kr_id]:
                if map_id not in groups.keys():
                    groups[map_id] = set()
                groups[map_id].add(r)
                
            # insert mds
            for md_id in kr_to_mds[kr_id]:
                if md_id not in groups.keys():
                    groups[md_id] = set()
                groups[md_id].add(r)
                
    # finally insert groups
    for group_id in groups.keys():
                
        # get group name
        if group_id.startswith('map'):
            name = map_to_name[group_id]
        if group_id.startswith('M'):
            name = md_to_name[group_id]
                    
        actual_group = cobra.core.Group(
            group_id, 
            name = name,
            members = list(groups[group_id]),
            kind = 'partonomy',
        )
        model.add_groups([actual_group])
        
        
        
    # insert custom groups:
    custom_groups = {
        'gr_ptdSTA': ['UAMAGLL', 'UMNS', 'UPPNAPT', 'UAGPT2', 'UAGPGAL', 'UAGPN6GT', 'UAAGGTLGAGT', 'UAAGGTLG3AGT', 'PPTGP'],
        'gr_HemeO': ['HEMEOS'],
        'gr_WTA1': ['ACGAMT', 'UNDBD', 'WTAGPT', 'WTAGPP', 'WTAUGLCT2', 'WTAALAT3', 'WTAPL3'],
        'gr_WTA2': ['ACGAMT', 'UNDBD', 'WTAGPT', 'WTARBT2', 'WTARPP2', 'WTAUGLCT', 'WTAALAT2', 'WTAPL2'],
        'gr_WTA3': ['ACGAMT', 'UNDBD', 'WTAGPT', 'WTAGPT2', 'WTARPP', 'WTAGLCNACT', 'WTAALAT', 'WTAPL'],
        'gr_LTA1': ['UGDIAT', 'UGLDIAT', 'GGGDAGF2', 'LIPOPO2', 'LTANACT', 'LTAALAT2'],
        'gr_LTA2': ['UGDIAT', 'UGADIAT', 'GGGDAGF', 'LIPOPO', 'LTAGAT', 'LTAALAT'],
    }
    for group_id in custom_groups.keys():
        actual_group = cobra.core.Group(
            group_id, 
            name = group_id,
            members = [model.reactions.get_by_id(rid) for rid in custom_groups[group_id]],
            kind = 'partonomy',
        )
        model.add_groups([actual_group])
        
        
    return model


    
def check_biosynthesis(logger, model, outdir, growth, biosynth):
    
    
    if growth: 
        
        # check production of biomass precursors: 
        logger.info("Checking biosynthesis of every biomass component...")
        
        print()
        mids = gempipe.check_reactants(model, 'Biomass')
        if mids == []: 
            print("No blocked biomass component detected!")
        print()

        
        
    if biosynth != '-':
        
        # check biosynthesis of every modeled metabolite:
        logger.info("Checking biosynthesis of every metabolite...")
        df_rows = []
        for m in model.metabolites:
            if m.id.endswith('_c'):
                binary, obj_value, status = gempipe.can_synth(model, m.id)
                df_rows.append({'mid': m.id, 'binary': binary, 'obj_value': obj_value, 'status': status})
        df_rows = pnd.DataFrame.from_records(df_rows)
        df_rows = df_rows.set_index('mid', drop=True, verify_integrity=True)
        
        # save table as excel: 
        df_rows.to_excel(f'{outdir}/biosynth.xlsx')
        logger.info(f"'{outdir}/biosynth.xlsx' created!")
        
        
        
        # focus on a particular metabolite:
        modeld_mids = [m.id for m in model.metabolites]
        if not (biosynth in modeld_mids and biosynth.endswith('_c')):
            logger.error(f"Cytosolic metabolite defined with --biosynth is not included: '{biosynth}'.")
            return 1
        
        nsol = 5   # number of solutions
        logger.info(f"Computing {nsol} gapfilling solutions for cytosolic metabolite {biosynth}...")
        gramneg = gempipe.get_universe(staining='neg')
        with model, gramneg:
            
            # mirror the medium
            gempipe.reset_growth_env(gramneg)
            for mid in ['glc__D_e', 'nh4_e', 'pi_e', 'so4_e', 'h2o_e', 'h_e', 'o2_e', 'co2_e']:
                gramneg.reactions.get_by_id(f"EX_{mid}").lower_bound = -1000
            for exr in ['EX_k_e', 'EX_mg2_e', 'EX_fe3_e', 'EX_cl_e', 'EX_ca2_e', 'EX_mn2_e', 'EX_cobalt2_e', 'EX_cu2_e', 'EX_zn2_e']:
                gramneg.reactions.get_by_id(exr).lower_bound = -1000   
            
            print()
            # perform gap-filling, solutions are shown using print()
            _ = gempipe.perform_gapfilling(model, gramneg, biosynth, nsol=nsol)
            print()
             
    
    return 0

  
    
def parse_eggnog(model, eggnog, idcollection_dict):
    
    
    eggnog = pnd.read_csv(eggnog, sep='\t', comment='#', header=None)
    eggnog.columns = 'query	seed_ortholog	evalue	score	eggNOG_OGs	max_annot_lvl	COG_category	Description	Preferred_name	GOs	EC	KEGG_ko	KEGG_Pathway	KEGG_Module	KEGG_Reaction	KEGG_rclass	BRITE	KEGG_TC	CAZy	BiGG_Reaction	PFAMs'.split('\t')
    eggnog = eggnog.set_index('query', drop=True, verify_integrity=True)
    
    
    # PART 1. get KO codes available
    kos_org = set()
    for gid, kos in eggnog['KEGG_ko'].items():
        if kos == '-': 
            continue
        kos = kos.split(',')
        kos = [i.replace('ko:', '') for i in kos]
        for ko in kos: 
            kos_org.add(ko)
            
            
    # PART 2. get reactions in the organism (even the GPR is not complete)
    kr_to_kos = idcollection_dict['kr_to_kos']
    krs_org = set()
    for kr, kos in kr_to_kos.items(): 
        if any([ko in kos_org for ko in kos]):
            krs_org.add(kr)
    
    
    return krs_org
    
    
    
def check_completeness(logger, model, progress, module, focus, eggnog, zeroes, idcollection_dict, summary_dict): 
    # check KEGG annotations in the universe model to get '%' of completeness per pathway/module.
    
            
            
    # get the reference set of kr codes (all kegg or organism specific): 
    if eggnog != '-':
        kr_uni = parse_eggnog(model, eggnog)
        kr_uni_label = "'eggnog annotation'"
    else: 
        kr_uni = idcollection_dict['kr']
        kr_uni_label = "'whole KEGG'"
    
    
    # get all the 'kr' annotations in the model
    kr_ids_modeled = set()
    for r in model.reactions: 
        if 'kegg.reaction' in r.annotation.keys():
            for kr_id in r.annotation['kegg.reaction']:
                kr_ids_modeled.add(kr_id)
    logger.info(f"Universe coverage for {kr_uni_label}: {round(len(kr_ids_modeled.intersection(kr_uni))/len(kr_uni)*100, 0)}%!")
    
    
    # get all the map / md codes:
    map_ids = set()
    md_ids = set()
    for i in summary_dict:
        map_ids.add(i['map_id'])
        for j in i['mds']:
            md_ids.add(j['md_id'])
            
            
    # check if 'focus' exist
    if focus != '-' and focus not in map_ids and focus not in md_ids:
        logger.error(f"The ID provided with --focus does not exist: {focus}.")
        return 1
    if focus.startswith('map'):
        logger.debug(f"With --focus {focus}, --module will switch to False.")
        module = False
    if focus != '-':
        missing_logger = ()
    
    
    # define some counters:
    maps_finished = set()
    maps_noreac = set()
    maps_missing = set()
    maps_partial = set()

    
    list_coverage  = []
    
    
    # iterate over each map:
    for i in summary_dict:
        
        
        # get ID and name: 
        map_id = i['map_id']
        map_name_short = f"{i['map_name'][:20]}"
        if len(i['map_name']) > 20: 
            map_name_short = map_name_short + '...'
        else:  # add spaces as needed: 
            map_name_short = map_name_short + ''.join([' ' for i in range(23-len(map_name_short))])
            
            
        # check if this map was (at least partially) covered:
        map_krs = set([kr for kr in i['kr_ids'] if kr in kr_uni])
        missing = map_krs - kr_ids_modeled
        present = kr_ids_modeled.intersection(map_krs)
        if focus == map_id: 
            missing_logger = (map_id, missing)

        
        if missing == set() and map_krs != set():
            maps_finished.add(map_id)
            
        elif map_krs == set():
            maps_noreac.add(map_id)
            
        elif missing == map_krs:
            maps_missing.add(map_id)
            
            if zeroes:
                list_coverage.append({
                    'map_id': map_id,
                    'map_name_short': map_name_short, 
                    'perc_completeness': 0,
                    'perc_completeness_str': ' 0',
                    'present': present,
                    'missing': missing,
                    'md_ids': [j['md_id'] for j in i['mds']],
                })
            
        elif len(missing) < len(map_krs):
            maps_partial.add(map_id)
            
            # get '%' of completeness:
            perc_completeness = len(present)/len(map_krs)*100
            perc_completeness_str = str(round(perc_completeness))   # version to be printed
            if len(perc_completeness_str)==1: 
                perc_completeness_str = ' ' + perc_completeness_str
                
            list_coverage.append({
                'map_id': map_id,
                'map_name_short': map_name_short, 
                'perc_completeness': perc_completeness,
                'perc_completeness_str': perc_completeness_str,
                'present': present,
                'missing': missing,
                'md_ids': [j['md_id'] for j in i['mds']],
            })
                
            
    # order list by '%' of completness and print:
    list_coverage = sorted(list_coverage, key=lambda x: x['perc_completeness'], reverse=True)
    for i in list_coverage:
        if progress:
            if focus=='-' or focus in i['md_ids'] or focus==i['map_id']:
                logger.info(f"{i['map_id']}: {i['map_name_short']} {i['perc_completeness_str']}% completed, {len(i['present'])} added, {len(i['missing'])} missing.")
        
        
        # get the correspondent pathway element of the 'summary_dict'
        right_item = None
        for k in summary_dict:
            if k['map_id'] == i['map_id']:
                right_item = k
                
                
        # define some counters:
        mds_completed = set()
        mds_noreac = set()
        mds_missing = set()
        mds_partial = set()


        list_coverage_md  = []
        spacer = '    '


        # iterate over each module:
        for z in right_item['mds']:


            # get ID and name: 
            md_id = z['md_id']
            md_name_short = f"{z['md_name'][:20]}"
            if len(z['md_name']) > 20: 
                md_name_short = md_name_short + '...'
            else:  # add spaces as needed: 
                md_name_short = md_name_short + ''.join([' ' for i in range(23-len(md_name_short))])


            # check if this module was (at least partially) covered:
            md_krs = set([kr for kr in z['kr_ids_md'] if kr in kr_uni])
            missing = md_krs - kr_ids_modeled
            present = kr_ids_modeled.intersection(md_krs)
            if focus == md_id: 
                missing_logger = (md_id, missing)
            
            
            if missing == set() and md_krs != set():
                mds_completed.add(md_id)

            elif md_krs == set():
                mds_noreac.add(md_id)

            elif missing == md_krs:
                mds_missing.add(md_id)
                
                if zeroes:
                    list_coverage_md.append({
                        'md_id': md_id,
                        'md_name_short': md_name_short, 
                        'perc_completeness': 0,
                        'perc_completeness_str': ' 0',
                        'present': present,
                        'missing': missing,
                    })
                
            elif len(missing) < len(md_krs):
                mds_partial.add(md_id)

                # get '%' of completeness:
                perc_completeness = len(present)/len(md_krs)*100
                perc_completeness_str = str(round(perc_completeness))   # version to be printed
                if len(perc_completeness_str)==1: 
                    perc_completeness_str = ' ' + perc_completeness_str

                list_coverage_md.append({
                    'md_id': md_id,
                    'md_name_short': md_name_short, 
                    'perc_completeness': perc_completeness,
                    'perc_completeness_str': perc_completeness_str,
                    'present': present,
                    'missing': missing,
                })
               
            
        # order list by '%' of completness and print:
        list_coverage_md = sorted(list_coverage_md, key=lambda x: x['perc_completeness'], reverse=True)
        for z in list_coverage_md:
            if module:
                if focus=='-' or focus==z['md_id']:
                    logger.info(f"{spacer}{z['md_id']}: {z['md_name_short']} {z['perc_completeness_str']}% completed, {len(z['present'])} added, {len(z['missing'])} missing.")
        
        
        # print summary:
        if module and focus=='-':
            logger.info(f"{spacer}Modules of {right_item['map_id']}: completed {len(mds_completed)} - partial {len(mds_partial)} - missing {len(mds_missing)} - noreac {len(mds_noreac)}")
    if focus != '-':
        logger.info(f"Missing reactions focusing on {missing_logger[0]}: {' '.join(list(missing_logger[1]))}.")
    logger.info(f"Maps: finished {len(maps_finished)} - partial {len(maps_partial)} - missing {len(maps_missing)} - noreac {len(maps_noreac)}")
            
        
    return 0     



def show_contributions(logger, db):
    
    # create a counter for each author
    cnt = {author: 0 for author in db['authors']['username']}
    cnt_tot = 0
    for rid, row in db['R'].iterrows():
        for author in row['author'].split(';'):
            author = author.rstrip().strip()
            cnt[author] += 1
            cnt_tot += 1
    for rid, row in db['T'].iterrows():
        for author in row['author'].split(';'):
            author = author.rstrip().strip()
            cnt[author] += 1
            cnt_tot += 1
        
    # compute percentages:
    pct = {author: cnt[author]/cnt_tot*100 for author in cnt.keys()}
    # sort in descending order: 
    pct = dict(sorted(pct.items(), key=lambda item: item[1], reverse=True))
    # convert to string:
    pct = {author: f'{round(pct[author],2)}%' for author in pct.keys()}
    logger.debug(f"Contributions: {pct}.")
        

    



import io
import requests

import pandas as pnd

import gempipe

__OPTTHR__ = 0.001   # optimization threshold



def force_id_on_sbml(file_path, model_id):
    
    with open(file_path, 'r') as file:
        content = file.read()
    content = content.replace(
        f'<model metaid="meta_{model_id}" fbc:strict="true">', 
        f'<model metaid="meta_{model_id}" id="{model_id}" fbc:strict="true">'
    )
    with open(file_path, 'w') as file:
        file.write(content)
        

def get_expcon(logger):
    
    
    logger.info("Downloading the experimental constraints file...")
    sheet_id = "1qGbIIipHJgYQjk3M0xDWKvnTkeInPoTeH9unDQkZPwg"
    url = f"https://docs.google.com/spreadsheets/d/{sheet_id}/export?format=xlsx"
    response = requests.get(url)  # download the requested file
    if response.status_code == 200:
        excel_data = io.BytesIO(response.content)   # load into memory
        exceldb = pnd.ExcelFile(excel_data)
    else:
        logger.error(f"Error during download. Please contact the developer.")
        return 1
    
    
    # check table presence
    sheet_names = exceldb.sheet_names
    for i in ['media', 'PM1', 'PM2A', 'PM3B', 'PM4A', 'authors']: 
        if i not in sheet_names:
            logger.error(f"Sheet '{i}' is missing!")
            return 1
        
        
    # load the tables
    expcon = {}
    expcon['media'] = exceldb.parse('media')
    expcon['PM1'] = exceldb.parse('PM1')
    expcon['PM2A'] = exceldb.parse('PM2A')
    expcon['PM3B'] = exceldb.parse('PM3B')
    expcon['PM4A'] = exceldb.parse('PM4A')
    expcon['authors'] = exceldb.parse('authors')
    
    
    # assign substrates as index
    expcon['media'].index = expcon['media'].iloc[:, 1]
    # remove first 2 useless column (empty & substrates)
    expcon['media'] = expcon['media'].iloc[:, 2:]
    
    
    for sheet in ['PM1', 'PM2A', 'PM3B', 'PM4A']:
        # assign wells as index
        expcon[sheet].index = expcon[sheet].iloc[:, 2]
        # remove first 3 useless columns
        expcon[sheet] = expcon[sheet].iloc[:, 3:]
    

    return expcon



def apply_medium_given_column(logger, model, medium, column, is_reference=False):
        
        
    # retrieve metadata
    description = column.iloc[0]
    doi = column.iloc[1]
    author = column.iloc[2]
    units = column.iloc[3]

    
    # convert substrates to dict
    column = column.iloc[4:]
    column = column.to_dict()

    
    # add trace elements:
    column['fe2'] = 'NL'
    column['mobd'] = 'NL'
    column['cobalt2'] = 'NL'


    # reset exchanges
    gempipe.reset_growth_env(model)    
    modeled_rids = [r.id for r in model.reactions]


    for substrate, value in column.items():

        if type(value)==float:
            continue   # empty cell, exchange will remain close

            
        # check if exchange is modeled
        if is_reference == False: 
            if f'EX_{substrate}_e' not in modeled_rids:
                logger.error(f"No exchange reaction found for substrate '{substrate}' in medium '{medium}'.")
                return 1
        else:  # external reference models might follow different standards.
            # The exr might not be present. 
            if f'EX_{substrate}_e' not in modeled_rids:
                logger.info(f"Reference has no exchange reaction for '{substrate}' in medium '{medium}': this substrate will be ignored.")
                continue


        # case "not limiting"
        value = value.strip().rstrip()
        if value == 'NL':   # non-limiting case
            model.reactions.get_by_id(f'EX_{substrate}_e').lower_bound = -1000


        # case "single value"
        elif '+-' not in value and '±' not in value:  # single number case
            value = value.replace(' ', '')  # eg "- 0.03" --> "-0.03"
            try: value = float(value)
            except: 
                logger.error(f"Invalid value found in medium '{medium}': '{substrate}' {value}.")
                return 1
            model.reactions.get_by_id(f'EX_{substrate}_e').lower_bound = value


        # case "with experimental error"
        else:  # value +- error
            if '±' in value: 
                value, error = value.split('±', 1)
            else: value, error = value.split('+-', 1)
            value = value.rstrip()
            error = error.strip()
            value = value.replace(' ', '')  # eg "- 0.03" --> "-0.03"
            try: value = float(value)
            except: 
                logger.error(f"Invalid value found in medium '{medium}': '{substrate}' {value} +- {error}.")
                return 1
            try: error = float(error)
            except: 
                logger.error(f"Invalid value found in medium '{medium}': '{substrate}' {value} +- {error}.")
                return 1
            model.reactions.get_by_id(f'EX_{substrate}_e').lower_bound = value -error
            model.reactions.get_by_id(f'EX_{substrate}_e').upper_bound = value +error

    return 0



def verify_growth(model, boolean=True):
        
    global __OPTTHR__
    res = model.optimize()
    obj_value = res.objective_value
    status = res.status
    if boolean:
        if obj_value < __OPTTHR__ or status=='infeasible':
            return False
        else: return True
    else:
        if status =='infeasible':
            return 'infeasible'
        elif obj_value < __OPTTHR__:
            return 0
        else:
            return round(obj_value, 3)
        
        
def log_metrics(logger, model, outmode='starting_uni'):
    
    
    G = len([g.id for g in model.genes])
    R = len([r.id for r in model.reactions if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites]))==1])
    T = len([r.id for r in model.reactions if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites]))!=1])
    M = len([m.id for m in model.metabolites])
    uM = len(set([m.id.rsplit('_',1)[0] for m in model.metabolites]))
    gr = len([gr.id for gr in model.groups])
    bP = len([m.id for m in model.reactions.get_by_id('Biomass').reactants])

    
    if   outmode == 'starting_uni':
        logger.info(f"Starting universe: [oG: {G}, R: {R}, T: {T}, uM: {uM}, bP: {bP}]")
    elif outmode == 'uni_features':
        biomass = round(model.slim_optimize(), 3)
        logger.info(f"Universe features: [oG: {G}, R: {R}, T: {T}, uM: {uM}, bP: {bP}, Biomass: {biomass}]")
    elif outmode == 'recon_model':
        logger.info(f"Resulting model: [G: {G}, R: {R}, T: {T}, uM: {uM}, bP: {bP}]")
    elif outmode == 'loaded_model':
        logger.info(f"Loaded model: [G: {G}, R: {R}, T: {T}, uM: {uM}, bP: {bP}]")



import time
import random
import os
import sys
import glob
import pickle
import copy


import pandas as pnd
from Bio.KEGG import REST
from Bio.KEGG.KGML import KGML_parser
#from Bio.Graphics.KGML_vis import KGMLCanvas   # <-- riquires reportlab





def download_raw_txtfiles(logger, outdir, usecache):
    

    # generate a README reporting the KEGG version number
    time.sleep(0.5)
    res_string = REST.kegg_info('kegg').read()
    RELEASE_kegg = copy.deepcopy(res_string)
    with open(f"{outdir}/kdown/RELEASE.txt", "w") as file:
        file.write(res_string)



    # get the list of the items included in the different databases: 
    databases = [
        'reaction',
        'compound',
        'enzyme',  # ec
        'orthology',
        'module', 
        'pathway',
    ]
    for db in databases:
        time.sleep(0.5)
        res_string = REST.kegg_list(db).read()
        with open(f"{outdir}/kdown/{db}.txt", "w") as file:
            file.write(res_string)



    # mix the items to download to be respectful/compliant
    items_to_download = []

    for db in databases:
        with open(f"{outdir}/kdown/{db}.txt", 'r') as file:
            res_string = file.read()
            rows = res_string.split('\n')
            for row in rows:
                item_id = row.split('\t', 1)[0]
                if item_id == '': continue
                items_to_download.append({'db': db, 'id': item_id})

    random.shuffle(items_to_download)



    # download the items
    for item in items_to_download:
        if usecache:
            if os.path.isfile(f"{outdir}/kdown/{item['db']}/{item['id']}.txt"):
                continue  # already downloaded, don't waste time
        time.sleep(0.5)
        try:
            res_string = REST.kegg_get(f"{item['db']}:{item['id']}").read()
        except: 
            logger.info(f"An error occurred during kegg_get of item {item['db']}:{item['id']}. Skipping the item.")
            continue
        os.makedirs(f"{outdir}/kdown/{item['db']}", exist_ok=True)
        with open(f"{outdir}/kdown/{item['db']}/{item['id']}.txt", "w") as file:
            file.write(res_string)
            
               
    return RELEASE_kegg



def create_dict_ko(logger, outdir):
    
    dict_ko = {}         # main output
    dict_headers = {}   # to track the available headers
    
    
    for file in glob.glob(f'{outdir}/kdown/orthology/K*.txt'):
        
        
        # create new item: 
        ko_id = file.rsplit('/', 1)[-1].replace('.txt', '')
        if ko_id not in dict_ko.keys():  
            dict_ko[ko_id] = {
                'kr_ids': set(),
                'symbols': set(),
                'name': None,
                'ecs': set(),
                'cogs': set(),
                'gos': set(),
            }   
        else:
            logger.error(f"{ko_id} already included!")
            return 1
        
        
        # read associated text file: 
        with open(file, 'r') as r_handler:
            file_content = r_handler.read().strip()
            
            
            # split rows:
            rows = file_content.split('\n')
            curr_header = None
            for row in rows:
                
                
                # track the headers
                header = row[:12]
                if header not in dict_headers.keys():
                    dict_headers[header] = {}
                    dict_headers[header]['cnt'] = 0
                    dict_headers[header]['ko_ids'] = set()
                dict_headers[header]['cnt'] += 1
                dict_headers[header]['ko_ids'].add(ko_id)
                
                
                # get the content
                if header == '///':
                    continue    # should be the end of the file
                if header != curr_header and header != '            ':
                    curr_header = header
                content = row[12:]
                
                
                # parse the entry: 
                if curr_header == 'ENTRY       ':
                    ko_id_content = content.strip().split(' ', 1)[0]
                    if ko_id_content != ko_id: 
                        logger.info(f"{ko_id} appears with a different ID in its file ({ko_id_content}).")
                        #return 1
                        
                        
                # parse the gene symbols
                if curr_header == 'SYMBOL      ':
                    symbols = content.strip().split(', ')
                    for symbol in symbols: 
                        if '.' in symbol:
                            continue 
                        dict_ko[ko_id]['symbols'].add(symbol)
                    # guarantee at least 1 symbol:
                    if dict_ko[ko_id]['symbols'] == set():
                        dict_ko[ko_id]['symbols'] = set([ko_id])
                        
                
                # parse the name / ECs
                if curr_header == 'NAME        ':
                    if '[' in content:
                        name = content.rsplit('[', 1)[0].strip()
                    else: name = content.strip()
                    dict_ko[ko_id]['name'] = name
                    if '[EC:' in content: 
                        ecs = content.rsplit('[EC:', 1)[1].replace(']','').split(' ')
                        for ec in ecs:
                            dict_ko[ko_id]['ecs'].add(ec)
                        
                        
                # parse the COG / GO terms
                if curr_header == 'DBLINKS     ':
                    if content.startswith('COG: '):
                        cogs = content[len('COG: '):].strip().split(' ')
                        for cog in cogs:
                            dict_ko[ko_id]['cogs'].add(cog)
                    if content.startswith('GO: '):
                        gos = content[len('GO: '):].strip().split(' ')
                        for go in gos:
                            dict_ko[ko_id]['gos'].add(go)
                        
                        
                # parse the reactions
                if curr_header == 'REACTION    ':
                    kr_ids = content.split(' ')
                    kr_ids = [i.strip() for i in kr_ids]
                    kr_ids = [i for i in kr_ids if i != '']
                    for kr_id in kr_ids: 
                        if len(kr_id)!=6 or kr_id[0]!='R':
                            #logger.info(f'{ko_id} is linked to a badly formatted reaction ID: {kr_id}.')
                            continue
                        else:
                            dict_ko[ko_id]['kr_ids'].add(kr_id)
                  
            
    logger.info(f'Number of unique items (K): {len(dict_ko.keys())}')
    logger.info('Metrics (K):\n\n'+ ''.join([f"'{header}': {dict_headers[header]['cnt']}\n" for header in dict_headers.keys()]))
    
    
    return dict_ko



def create_dict_c(logger, outdir):
    
    dict_c = {}         # main output
    dict_headers = {}   # to track the available headers
    
    
    for file in glob.glob(f'{outdir}/kdown/compound/C*.txt'):
        
        
        # create new item: 
        kc_id = file.rsplit('/', 1)[-1].replace('.txt', '')
        if kc_id not in dict_c.keys():  
            dict_c[kc_id] = {
                'kr_ids': set(), 
                'map_ids': set(), 
                'md_ids': set()
            }   
        else:
            logger.error(f"{kc_id} already included!")
            return 1
        
        
        # read associated text file: 
        with open(file, 'r') as r_handler:
            file_content = r_handler.read().strip()
            
            
            # split rows:
            rows = file_content.split('\n')
            curr_header = None
            for row in rows:
                
                
                # track the headers
                header = row[:12]
                if header not in dict_headers.keys():
                    dict_headers[header] = {}
                    dict_headers[header]['cnt'] = 0
                    dict_headers[header]['kc_ids'] = set()
                dict_headers[header]['cnt'] += 1
                dict_headers[header]['kc_ids'].add(kc_id)
                
                
                # get the content
                if header == '///':
                    continue    # should be the end of the file
                if header != curr_header and header != '            ':
                    curr_header = header
                content = row[12:]
                
                
                # parse the entry: 
                if curr_header == 'ENTRY       ':
                    kc_id_content = content.strip().split(' ', 1)[0]
                    if kc_id_content != kc_id: 
                        logger.info(f"{kc_id} appears with a different ID in its file ({kc_id_content}).")
                        #return 1
                        
                        
                # parse the reactions
                if curr_header == 'REACTION    ':
                    kr_ids = content.split(' ')
                    kr_ids = [i.strip() for i in kr_ids]
                    kr_ids = [i for i in kr_ids if i != '']
                    for kr_id in kr_ids: 
                        if len(kr_id)!=6 or kr_id[0]!='R':
                            logger.info(f'{kc_id} is linked to a badly formatted reaction ID: {kr_id}.')
                        else:
                            dict_c[kc_id]['kr_ids'].add(kr_id)
                        
                
                # parse maps (pathways)
                if curr_header == 'PATHWAY     ':
                    map_id = content.strip().split(' ')[0]    # one per row
                    if len(map_id)!=8 or map_id[:3]!='map':
                        # handle cases like eg "map01100(G00003)" (C05860)
                        if len(map_id)!=16 or map_id[8:10]!='(G' or map_id[15]!=')':
                            logger.info(f'{kc_id} is linked to a badly formatted pathway ID: {map_id}.')
                        else:
                            map_id = map_id[:8]
                            dict_c[kc_id]['map_ids'].add(map_id)
                    else:
                        dict_c[kc_id]['map_ids'].add(map_id)
                        
                            
                # parse modules 
                if curr_header == 'MODULE      ':
                    md_id = content.strip().split(' ')[0]    # one per row
                    if len(md_id)!=6 or md_id[0]!='M':
                        logger.info(f'{kc_id} is linked to a badly formatted module ID: {md_id}.')
                    else:
                        dict_c[kc_id]['md_ids'].add(md_id)


    logger.info(f'Number of unique items (C): {len(dict_c.keys())}')
    logger.info('Metrics (C):\n\n'+ ''.join([f"'{header}': {dict_headers[header]['cnt']}\n" for header in dict_headers.keys()]))
    
    
    return dict_c



def create_dict_r(logger, outdir):
    
    dict_r = {}         # main output
    dict_headers = {}   # to track the available headers
    
    
    for file in glob.glob(f'{outdir}/kdown/reaction/R*.txt'):
        
        
        # create new item: 
        kr_id = file.rsplit('/', 1)[-1].replace('.txt', '')
        if kr_id not in dict_r.keys():  
            dict_r[kr_id] = {
                'map_ids': set(), 
                'md_ids': set(), 
                'ko_ids': set()
            }  
        else:
            logger.error(f"{kr_id} already included!")
            return 1
        
        
        # read associated text file: 
        with open(file, 'r') as r_handler:
            file_content = r_handler.read().strip()
            
            
            # split rows:
            rows = file_content.split('\n')
            curr_header = None
            for row in rows:
                
                
                # track the headers
                header = row[:12]
                if header not in dict_headers.keys():
                    dict_headers[header] = {}
                    dict_headers[header]['cnt'] = 0
                    dict_headers[header]['kr_ids'] = set()
                dict_headers[header]['cnt'] += 1
                dict_headers[header]['kr_ids'].add(kr_id)
                
                  
                # get the content
                if header == '///':
                    continue    # should be the end of the file
                if header != curr_header and header != '            ':
                    curr_header = header
                content = row[12:]
                
                
                # parse the entry: 
                if curr_header == 'ENTRY       ':
                    kr_id_content = content.strip().split(' ', 1)[0]
                    if kr_id_content != kr_id: 
                        logger.info(f"{kr_id} appears with a different ID in its file ({kr_id_content}).")
                        #return 1
                        
                
                # parse maps (pathways)
                if curr_header == 'PATHWAY     ':
                    map_id = content.strip().split(' ')[0]    # one per row
                    if len(map_id)!=7 or map_id[:2]!='rn':
                        logger.info(f'{kr_id} is linked to a badly formatted pathway ID: {map_id}.')
                    else:
                        map_id = 'map' + map_id[2:]
                        dict_r[kr_id]['map_ids'].add(map_id)
                        
                        
                # parse modules 
                if curr_header == 'MODULE      ':
                    md_id = content.strip().split(' ')[0]    # one per row
                    if len(md_id)!=6 or md_id[0]!='M':
                        logger.info(f'{kr_id} is linked to a badly formatted module ID: {md_id}.')
                    else:
                        dict_r[kr_id]['md_ids'].add(md_id)
                        
                        
                # parse orthologs 
                if curr_header == 'ORTHOLOGY   ':
                    ko_id = content.strip().split(' ')[0]    # one per row
                    if len(ko_id)!=6 or ko_id[0]!='K':
                        logger.info(f'{kr_id} is linked to a badly formatted ortholog ID: {ko_id}.')
                    else:
                        dict_r[kr_id]['ko_ids'].add(ko_id)

    
    logger.info(f'Number of unique items (R): {len(dict_r.keys())}')
    logger.info('Metrics (R):\n\n'+ ''.join([f"'{header}': {dict_headers[header]['cnt']}\n" for header in dict_headers.keys()]))
    
    
    return dict_r



def create_dict_map(logger, outdir):
    
    dict_map = {}         # main output
    dict_headers = {}   # to track the available headers
    
    
    for file in glob.glob(f'{outdir}/kdown/pathway/map*.txt'):
        
        
        # create new item: 
        map_id = file.rsplit('/', 1)[-1].replace('.txt', '')
        if map_id not in dict_map.keys():  
            dict_map[map_id] = {
                'md_ids': set(), 
                'name': None,
            }   
        else:
            logger.error(f"{map_id} already included!")
            return 1
        
        
        # read associated text file: 
        with open(file, 'r') as r_handler:
            file_content = r_handler.read().strip()
            
            
            # split rows:
            rows = file_content.split('\n')
            curr_header = None
            for row in rows:
                
                
                # track the headers
                header = row[:12]
                if header not in dict_headers.keys():
                    dict_headers[header] = {}
                    dict_headers[header]['cnt'] = 0
                    dict_headers[header]['map_ids'] = set()
                dict_headers[header]['cnt'] += 1
                dict_headers[header]['map_ids'].add(map_id)
                
                
                # get the content
                if header == '///':
                    continue    # should be the end of the file
                if header != curr_header and header != '            ':
                    curr_header = header
                content = row[12:]
                
                
                # parse the entry: 
                if curr_header == 'ENTRY       ':
                    map_id_content = content.strip().split(' ', 1)[0]
                    if map_id_content != map_id: 
                        logger.info(f"{map_id} appears with a different ID in its file ({map_id_content}).")
                        #return 1
                        
                        
                # parse the name:
                if curr_header == 'NAME        ': 
                    name = content.strip()
                    dict_map[map_id]['name'] = name

                        
                # parse modules 
                if curr_header == 'MODULE      ':
                    md_id = content.strip().split(' ')[0]    # one per row
                    if len(md_id)!=6 or md_id[0]!='M':
                        logger.info(f'{map_id} is linked to a badly formatted module ID: {md_id}.')
                    else:
                        dict_map[map_id]['md_ids'].add(md_id)

    
    logger.info(f'Number of unique items (map): {len(dict_map.keys())}')
    logger.info('Metrics (map):\n\n'+ ''.join([f"'{header}': {dict_headers[header]['cnt']}\n" for header in dict_headers.keys()]))
    
    
    return dict_map



def create_dict_md(logger, outdir):
    
    dict_md = {}         # main output
    dict_headers = {}   # to track the available headers
    
    
    for file in glob.glob(f'{outdir}/kdown/module/M*.txt'):
        
        
        # create new item: 
        md_id = file.rsplit('/', 1)[-1].replace('.txt', '')
        if md_id not in dict_md.keys():  
            dict_md[md_id] = {
                'name': None,
            }   
        else:
            logger.error(f"{md_id} already included!")
            return 1
        
        
        # read associated text file: 
        with open(file, 'r') as r_handler:
            file_content = r_handler.read().strip()
            
            
            # split rows:
            rows = file_content.split('\n')
            curr_header = None
            for row in rows:
                
                
                # track the headers
                header = row[:12]
                if header not in dict_headers.keys():
                    dict_headers[header] = {}
                    dict_headers[header]['cnt'] = 0
                    dict_headers[header]['md_ids'] = set()
                dict_headers[header]['cnt'] += 1
                dict_headers[header]['md_ids'].add(md_id)
                
                
                # get the content
                if header == '///':
                    continue    # should be the end of the file
                if header != curr_header and header != '            ':
                    curr_header = header
                content = row[12:]
                
                
                # parse the entry: 
                if curr_header == 'ENTRY       ':
                    md_id_content = content.strip().split(' ', 1)[0]
                    if md_id_content != md_id: 
                        logger.info(f"{md_id} appears with a different ID in its file ({md_id_content}).")
                        #return 1
                        
                        
                # parse the name:
                if curr_header == 'NAME        ': 
                    name = content.strip()
                    dict_md[md_id]['name'] = name

    
    logger.info(f'Number of unique items (M): {len(dict_md.keys())}')
    logger.info('Metrics (M):\n\n'+ ''.join([f"'{header}': {dict_headers[header]['cnt']}\n" for header in dict_headers.keys()]))
    
    
    return dict_md



def create_idcollection_dict(dict_ko, dict_c, dict_r, dict_map, dict_md):

    idcollection_dict = {}
    
    
    idcollection_dict['kc'] = set()
    for kc_id in dict_c.keys():
        idcollection_dict['kc'].add(kc_id)
        
        
    idcollection_dict['kr'] = set()
    for kr_id in dict_r.keys():
        idcollection_dict['kr'].add(kr_id)
        
        
    idcollection_dict['ko'] = set()
    for ko_id in dict_ko.keys():
        idcollection_dict['ko'].add(ko_id)
        
        
    idcollection_dict['kr_to_kos'] = {}
    for kr_id in dict_r.keys():
        idcollection_dict['kr_to_kos'][kr_id] = set()
        for ko_id in dict_r[kr_id]['ko_ids']:
            idcollection_dict['kr_to_kos'][kr_id].add(ko_id)
            
            
    idcollection_dict['kr_to_maps'] = {}
    for kr_id in dict_r.keys():
        idcollection_dict['kr_to_maps'][kr_id] = set()
        for map_id in dict_r[kr_id]['map_ids']:
            idcollection_dict['kr_to_maps'][kr_id].add(map_id)
            
            
    idcollection_dict['kr_to_mds'] = {}
    for kr_id in dict_r.keys():
        idcollection_dict['kr_to_mds'][kr_id] = set()
        for md_id in dict_r[kr_id]['md_ids']:
            idcollection_dict['kr_to_mds'][kr_id].add(md_id)
            
    
    idcollection_dict['ko_to_name'] = {}
    for ko_id in dict_ko.keys():
        idcollection_dict['ko_to_name'][ko_id] = dict_ko[ko_id]['name']
    
    
    idcollection_dict['ko_to_symbols'] = {}
    for ko_id in dict_ko.keys():
        idcollection_dict['ko_to_symbols'][ko_id] = set()
        for symbol in dict_ko[ko_id]['symbols']:
            idcollection_dict['ko_to_symbols'][ko_id].add(symbol)
            
            
    idcollection_dict['ko_to_ecs'] = {}
    for ko_id in dict_ko.keys():
        idcollection_dict['ko_to_ecs'][ko_id] = set()
        for ec in dict_ko[ko_id]['ecs']:
            idcollection_dict['ko_to_ecs'][ko_id].add(ec)
            
            
    idcollection_dict['ko_to_cogs'] = {}
    for ko_id in dict_ko.keys():
        idcollection_dict['ko_to_cogs'][ko_id] = set()
        for cog in dict_ko[ko_id]['cogs']:
            idcollection_dict['ko_to_cogs'][ko_id].add(cog)
            
            
    idcollection_dict['ko_to_gos'] = {}
    for ko_id in dict_ko.keys():
        idcollection_dict['ko_to_gos'][ko_id] = set()
        for go in dict_ko[ko_id]['gos']:
            idcollection_dict['ko_to_gos'][ko_id].add(go)
            
            
    idcollection_dict['map_to_name'] = {}
    for map_id in dict_map.keys():
        idcollection_dict['map_to_name'][map_id] = dict_map[map_id]['name']
        
        
    idcollection_dict['md_to_name'] = {}
    for md_id in dict_md.keys():
        idcollection_dict['md_to_name'][md_id] = dict_md[md_id]['name']
        
    
    return idcollection_dict



def create_summary_dict(dict_c, dict_r, dict_map, dict_md):

    summary_dict = []   # list fo dicts

    for i, map_id in enumerate(list(dict_map.keys())):
        
        
        # get total number of reactions in this map.
        # iterate dict_r looking at the linked maps for each reaction 
        kr_ids = set()
        for r_id in dict_r.keys():
            if map_id in dict_r[r_id]['map_ids']:
                kr_ids.add(r_id)
        
        
        # append the new elemnt (dict) to the list
        summary_dict.append({})
        summary_dict[i]['map_id'] = map_id
        summary_dict[i]['map_name'] = dict_map[map_id]['name']
        summary_dict[i]['kr_ids'] = kr_ids
        summary_dict[i]['cnt_r'] = len(kr_ids)
        summary_dict[i]['mds'] = []
        
        
        # get number of reactions per module involved.
        # iterate dict_r looking at the linked modules for each reaction 
        for k, md_id in enumerate(list(dict_map[map_id]['md_ids'])):
            kr_ids_md = set()
            for r_id in dict_r.keys():
                if md_id in dict_r[r_id]['md_ids']:
                    kr_ids_md.add(r_id)
            
            summary_dict[i]['mds'].append({})
            summary_dict[i]['mds'][k]['md_id'] = md_id
            summary_dict[i]['mds'][k]['md_name'] = dict_md[md_id]['name']
            summary_dict[i]['mds'][k]['kr_ids_md'] = kr_ids_md
            summary_dict[i]['mds'][k]['cnt_r_md'] = len(kr_ids_md)
        summary_dict[i]['mds'] = sorted(summary_dict[i]['mds'], key=lambda x: x['cnt_r_md'], reverse=True)
                    
          
    summary_dict = sorted(summary_dict, key=lambda x: x['cnt_r'], reverse=True)
    
    
    return summary_dict


import pandas as pnd
import cobra



def get_deprecated_kos():
    deprecated_kos = [
        'K11189',  # should be K02784
    ]
    return deprecated_kos



def check_gpr(logger, rid, row, kr_ids, idcollection_dict, addtype='R'): 
    
    itemtype = 'Reaction' if addtype=='R' else 'Transporter'
    
    # check presence of the GPR
    if pnd.isna(row['gpr_manual']): 
        logger.error(f"{itemtype} '{rid}' has missing GPR: '{row['gpr_manual']}'.")
        return 1
    
    # get ko_ids in this reaction:
    ko_ids_parsed = row['gpr_manual'].strip()
    ko_ids_parsed = ko_ids_parsed.replace(' and ', ',')
    ko_ids_parsed = ko_ids_parsed.replace(' or ', ',')
    ko_ids_parsed = ko_ids_parsed.replace('(', '')
    ko_ids_parsed = ko_ids_parsed.replace(')', '')
    ko_ids_parsed = ko_ids_parsed.split(',')
    
    
    # collect all the ko_ids for this reaction:
    if addtype=='R':
        ko_for_rid = set()   
        for kr_id in kr_ids:
            if kr_id == 'RXXXXX': 
                continue
            ko_for_rid = ko_for_rid.union(idcollection_dict['kr_to_kos'][kr_id])
            
    
    # check if these ko_ids exist:
    for ko_id in ko_ids_parsed:
        ko_id = ko_id.strip() 
        
        if ko_id in get_deprecated_kos():
            pass
        elif ko_id not in idcollection_dict['ko'] and ko_id != 'spontaneous' and ko_id != 'orphan':
            logger.error(f"{itemtype} '{rid}' has an invalid KEGG Ortholog: '{ko_id}'.")
            return 1
        
        
        # check if these ko_ids are really assigned to this reaction:
        if addtype=='R':
            if ko_id not in ko_for_rid and ko_id != 'spontaneous' and ko_id != 'orphan':
                if rid not in [  # verified manual corrections
                    'SUCD1', 'ALKP', 'PFK_3', 'TCMPTS', 'PPA', 'APSR',
                    'NADHPO', 'ACOD', '5DOAN', '5DRBK', '5DRBPI', '5DRBUPD',
                    'FNPOR', 'THZPSN', 'PPTGP'
                ]:
                    logger.debug(f"Ortholog '{ko_id}' should not be linked to reaction '{rid}' (available for {kr_ids}: {ko_for_rid}).")
            
    
    # check if some ko_ids are missing from this reaction:
    if addtype=='R':
        missing_ko_ids = ko_for_rid - (set(ko_ids_parsed) - set(['spontaneous', 'orphan']))
        if len(missing_ko_ids) > 0:
            logger.error(f"Orthologs {missing_ko_ids} are missing from reaction '{rid}' ({kr_ids}).")
            return 1
            
            
    return 0
                


def add_reaction(logger, model, rid, row, kr_ids, addtype='R'):
    
    
    itemtype = 'Reaction' if addtype=='R' else 'Transporter'
    
    
    r = cobra.Reaction(rid)
    model.add_reactions([r])
    r = model.reactions.get_by_id(rid)
    r.build_reaction_from_string(row['rstring'])
    r.name = row['name'].strip()
    r.bounds = (0, 1000) if ' --> ' in row['rstring'] else (-1000,  1000)
    r.gene_reaction_rule = row['gpr_manual'].strip()
    r.update_genes_from_gpr()
    for m in r.metabolites:
        if m.formula == None or m.charge == None:
            logger.error(f"Metabolite '{m.id}' appears in '{r.id}' but was not previously defined.")
            return 1
        
        
    # add kr annotations to model
    r.annotation['kegg.reaction'] = kr_ids
        
        
    # check if unbalanced
    if r.check_mass_balance() != {}: 
        logger.error(f"{itemtype} '{r.id}' is unbalanced: {r.check_mass_balance()}.")
        return 1    
    
    
    return 0



def check_rstring_arrow(logger, row, addtype='R'):
    
    itemtype = 'Reaction' if addtype=='R' else 'Transporter'
    
    
    if pnd.isna(row['rstring']): 
        logger.error(f"{itemtype} '{rid}' has no definition (rstring).")
        return 1
    if ' --> ' not in row['rstring'] and ' <=> ' not in row['rstring']:
        logger.error(f"{itemtype} '{rid}' has invalid arrow: '{row['rstring']}'.")
        return 1
        
        
    return 0



def check_author(logger, mrid, row, db, addtype='R'): 
    
    
    if addtype=='M':
        itemtype = 'Metabolite'
    elif  addtype=='R' :
        itemtype = 'Reaction'
    else: itemtype = 'Transporter'
    
    
    if pnd.isna(row['author']): 
        logger.error(f"{itemtype} '{mrid}' has no author.")
        return 1
    
    authors = set()
    for author in row['author'].split(';'):
        author = author.strip()
        authors.add(author)
        if author not in db['authors']['username'].to_list(): 
            logger.error(f"{itemtype} '{mrid}' has invalid author: '{author}'.")
            return 1
    
    
    return list(authors)
    
    
    



import os
import io
import requests
import logging
import warnings
from pathlib import Path
from importlib import resources


import pandas as pnd
import cobra


import gempipe

from ..commons import apply_medium_given_column
from ..commons import verify_growth
from ..commons import log_metrics

__OPTTHR__ = 0.001   # optimization threshold





def get_biomass_dict():
    
    
    # Note: universal and conditional precursors have been defined in Xavier2017 .
    # Here presented in the same order of Xavier2017 .
    # Xavier2017: 10.1016/j.ymben.2016.12.002.
    biomass_dict = {
        ##### UNIVERSAL #####
        'ribo_nucleotides': [
            'atp_c', 'ctp_c', 'gtp_c', 'utp_c'
        ],
        'deoxyribo_nucleotides': [
            'datp_c', 'dctp_c', 'dgtp_c', 'dttp_c'
        ],
        'amino_acids': [
            'ala__L_c', 'arg__L_c', 'asn__L_c', 'asp__L_c', 'cys__L_c', 
            'gln__L_c', 'glu__L_c', 'gly_c',    'his__L_c', 'ile__L_c', 
            'leu__L_c', 'lys__L_c', 'met__L_c', 'ser__L_c', 'pro__L_c', 
            'thr__L_c', 'trp__L_c', 'tyr__L_c', 'val__L_c', 'phe__L_c'
        ],
        'cofactors_uni': [
            'nad_c',    # B3: Nicotinamide -adenine dinucleotide phosphate
            'nadp_c',   # B3: Nicotinamide -adenine dinucleotide phosphate
            'coa_c',    # B5: Coenzyme A  (dependant on 'pnto__R_c')
            'fad_c',    # B2: Flavin adenine dinucleotide
            'fmn_c',    # B2: Flavin mononucleotide
            'ribflv_c', # B2: ribovlavin. Non-active form acording to Xavier2017 but anyway included.
            #'f4200_c', # B2: included by Xavier2017 in 'universal' but the description seems conditional.
            'thf_c',    # B9: tetrahydrofolate 
            '10fthf_c', # B9: 10-Formyltetrahydrofolate
            '5mthf_c',  # B9: 5-Methyltetrahydrofolate
            'thmpp_c',  # B1: Thiamine diphosphate
            'pydx5p_c', # B6: pyridoxal 5-phosphate
            'amet_c',   # SAM: S-adenosyl-methionine
        ],
        ##### CONDITIONAL #####
        'cofactors_con': [
            #'f4200_c', # coenzyme f420 (electron transfer in methanogens, actinobacteria , and others)
            'ptrc_c',   # Putrescine
            'spmd_c',   # Sperimidine
            'pheme_c',  # protoheme
            'mql8_c',   # menaquinol / manaquinone (mqn8_c)
            'q8h2_c',   # ubiquinol / ubiquinone (q8_c)
            # Methionaquinone
            'btn_c',    # B7: biotin
            'ACP_c',    # Acyl-carrier protein
            'adocbl_c', # B12: Adenosylcobalamin
            # Lipoate
            'uacgam_c'  # uridine diphosphate N-Acetylglucosamine (UDP-GlcNAc)
        ],
        ##### ADDED ##### (conditionals not included or lumped in Xavier2017)
        'cofactors_add': [
            'hemeO_c',  # heme-O
            'sheme_c',  # siroheme
            'moco_c',   # molybdenum cofactor
            'phllqol_c',# phylloquinol / phylloquinone (phllqne_c)
            'gthrd_c',  # glutathione (reduced)
        ],
        'lipids': [
            'pe120_c', # phosphatidyl-ethanolamine (12:0;12:0)
            'pg120_c', # phosphatidyl-glycerol (12:0;12:0)
            'clpn120_c', # cardiolipin (12:0;12:0;12:0;12:0)
            # 1-lysyl phosphatidylglycerol (plantarum)
        ],
        'membrane_wall': [
            'peptidoSTA_c', # peptidoglycan (dependant on 'udcpdp_c')
            'WTAgg40r_20n_20a_P_c', # teichoic acids
            'WTArg40r_20g_20a_P_c', # teichoic acids
            'WTAg40g_20g_20a_P_c', # teichoic acids
            'LTAgg40g_20n_20a_c', # lipoteichoic acids
            'LTAga40g_20t_20a_c', # lipoteichoic acids            
            # capsular polysaccharides
            # kdo_lipid_A
        ],
        'energy_stock': [
            # glycogen
            # starch
            # PHA / PHB
        ]
    }
    return biomass_dict

    
    
    
def check_inputs(logger, universe, eggnog):
    
    
    # check if files exist
    if os.path.isfile(universe) == False: 
        logger.error(f"Provided --universe doesn't exist: {universe}.")
        return 1
    if os.path.isfile(eggnog) == False: 
        logger.error(f"Provided --eggnog doesn't exist: {eggnog}.")
        return 1
    
    
    # check the universe model format
    if universe.endswith('.xml'):
        universe = cobra.io.read_sbml_model(universe)
    else: 
        logger.error(f"Provided --universe must be in cobrapy-compatible SBML format (.xml extension).")
        return 1
    
    
    # log main universe metrics:
    log_metrics(logger, universe, outmode='starting_uni')
        
        
    # load eggnog annotations
    eggnog = pnd.read_csv(eggnog, sep='\t', comment='#', header=None)
    eggnog.columns = 'query	seed_ortholog	evalue	score	eggNOG_OGs	max_annot_lvl	COG_category	Description	Preferred_name	GOs	EC	KEGG_ko	KEGG_Pathway	KEGG_Module	KEGG_Reaction	KEGG_rclass	BRITE	KEGG_TC	CAZy	BiGG_Reaction	PFAMs'.split('\t')
    eggnog = eggnog.set_index('query', drop=True, verify_integrity=True)
    
    return [universe, eggnog]



def parse_eggnog(eggnog):
    
    
    # PART 1. get KO codes available
    gid_to_kos = {}
    ko_to_gids = {}
    for gid, kos in eggnog['KEGG_ko'].items():
        if kos == '-': 
            continue
            
        if gid not in gid_to_kos.keys(): 
            gid_to_kos[gid] = set()
            
        kos = kos.split(',')
        kos = [i.replace('ko:', '') for i in kos]
        for ko in kos: 
            if ko not in ko_to_gids.keys(): 
                ko_to_gids[ko] = set()
                
            # populate dictionaries
            ko_to_gids[ko].add(gid)
            gid_to_kos[gid].add(ko)

    
    return ko_to_gids, gid_to_kos



def get_modeled_kos(model):
    
    
    # get modeled KO ids:
    modeled_gid_to_ko = {}
    modeled_ko_to_gid = {}
    
    for g in model.genes:
        if g.id in ['orphan', 'spontaneous']: 
            continue
        corresponding_ko = g.annotation['ko']
        
        modeled_gid_to_ko[g.id] = corresponding_ko
        modeled_ko_to_gid[corresponding_ko] = g.id
        
    modeled_kos = list(modeled_gid_to_ko.values())
        
    return modeled_kos, modeled_gid_to_ko, modeled_ko_to_gid



def subtract_kos(logger, model, eggonog_ko_to_gids):
    
    
    modeled_kos, _, modeled_ko_to_gid = get_modeled_kos(model)
        
        
    to_remove = []  # genes to delete
    for ko in modeled_kos: 
        if ko not in eggonog_ko_to_gids.keys():
            gid_to_remove = modeled_ko_to_gid[ko]
            to_remove.append(model.genes.get_by_id(gid_to_remove))
            
    
    # remove also orphan reactions!
    to_remove.append(model.genes.get_by_id('orphan'))
    
    
    # delete marked genes!
    # trick to avoid the WARNING "cobra/core/group.py:147: UserWarning: need to pass in a list" 
    # triggered when trying to remove reactions that are included in groups. 
    with warnings.catch_warnings():  # temporarily suppress warnings for this block
        warnings.simplefilter("ignore")  # ignore all warnings
        cobra_logger = logging.getLogger("cobra.util.solver")
        old_level = cobra_logger.level
        cobra_logger.setLevel(logging.ERROR)   

        cobra.manipulation.delete.remove_genes(model, to_remove, remove_reactions=True)

        # restore original behaviour: 
        cobra_logger.setLevel(old_level)   
        
   
    logger.info(f"Found {len(model.genes)} modeled orthologs.")
    return 0



def translate_remaining_kos(logger, model, eggonog_ko_to_gids):
    
    
    _, modeled_gid_to_ko, _ = get_modeled_kos(model) 
    
    
    # iterate reactions:
    for r in model.reactions:

        gpr = r.gene_reaction_rule

        # force each gid to be surrounded by spaces: 
        gpr = ' ' + gpr.replace('(', ' ( ').replace(')', ' ) ') + ' '
        
        for gid in modeled_gid_to_ko.keys():
            if f' {gid} ' in gpr:
                
                new_gids = eggonog_ko_to_gids[modeled_gid_to_ko[gid]]
                gpr = gpr.replace(f' {gid} ', f' ({" or ".join(new_gids)}) ')       
            

        # remove spaces between parenthesis
        gpr = gpr.replace(' ( ', '(').replace(' ) ', ')')
        # remove spaces at the extremes: 
        gpr = gpr[1: -1]


        # New genes are introduced. Parethesis at the extremes are removed if not necessary. 
        r.gene_reaction_rule = gpr
        r.update_genes_from_gpr()
            
            
    # remaining old 'Cluster_'s need to removed.
    # remove if (1) hte ID starts with clusters AND (2) they are no more associated with any reaction
    to_remove = []
    for g in model.genes:
        
        if g.id in ['orphan', 'spontaneous']:
            continue
            
        if g.id in modeled_gid_to_ko.keys() and len(g.reactions)==0:
            to_remove.append(g)
            
    # warning suppression not needed here, as no reaction is actually removed.
    cobra.manipulation.delete.remove_genes(model, to_remove, remove_reactions=True)
    
        
    logger.info(f"Translated orthologs to {len(model.genes)} genes.")
    return 0
        
    

def restore_gene_annotations(logger, model, universe, eggonog_gid_to_kos):
    
    
    for g in model.genes:
        if g.id == 'spontaneous': 
            continue
            
        names = []    
        for ko in eggonog_gid_to_kos[g.id]:
            
            # get the corresponding universal gene:
            uni_g = None
            for ug in universe.genes:
                if 'ko' not in ug.annotation.keys():
                    continue
                if ug.annotation['ko']==ko:  # take the first (and only)
                    uni_g = ug
                    break
            if uni_g == None:  
                # The ko provided by eggnog-mapper is still not modeled in the universe.
                # Multiple ko are possible for each gene. Of these, only 1 could b modeled.
                continue
            
            
            # transfer annotations of this ko/universal gene:
            for key in uni_g.annotation.keys():
                if key == 'ko':
                    continue   # resulting models will loose links to kos.
                if key not in g.annotation:
                    g.annotation[key] = []
                items = uni_g.annotation[key]
                if type(items)==str:  items = [items]
                for i in items:
                    g.annotation[key].append(i)
                    
            # collect names
            names.append(uni_g.name)
        g.name = '; '.join(names)


    
def include_forced(logger, model, universe, force_inclusion):
    
    if force_inclusion != '-':
        forced_rids = force_inclusion.split(',')
        
        modeled_rids = [r.id for r in model.reactions]
        universal_rids = [r.id for r in universe.reactions]
        
        introduced_rids = []
        for rid in forced_rids: 
            
            if rid not in universal_rids:
                logger.info(f"Ignoring reaction ID {rid} since it's not included in the universe.")
                continue
            
            if rid not in modeled_rids:
                gempipe.import_from_universe(model, universe, rid, gpr='')
                introduced_rids.append(rid)
            else:
                logger.debug(f"Requested reaction ID {rid} was already included.")
        logger.info(f"Reactions forcibly included and orphans: {introduced_rids}.")
        
    return 0



def adjust_biomass_equation(logger, model, universe):
    
    
    # Note: universal and conditional precursors have been defined in 10.1016/j.ymben.2016.12.002 .
    precursor_to_pathway = {  # listing alternative biosynthetic pathways
        
        ##### cofactors_con #####
        #'f4200_c', # coenzyme f420 (electron transfer in methanogens, actinobacteria , and others)
        #'ptrc_c',   # Putrescine
        #'spmd_c',   # Sperimidine
        'pheme_c': ['M00868', 'M00121', 'M00926'],   # protoheme (heme)
        'mql8_c': ['M00116', 'M00930', 'M00931'],  # menaquinol
        'q8h2_c': ['M00117', 'M00989', 'M00128'],  # ubiquinol
        # Methionaquinone
        #'btn_c',    # B7: biotin
        #'ACP_c',    # Acyl-carrier protein
        'adocbl_c': ['M00122'],   # vitamin B12 (cobolamin)
        # Lipoate
        #'uacgam_c'  # uridine diphosphate N-Acetylglucosamine (UDP-GlcNAc)
        
        ##### cofactors_add #####
        'hemeO_c': ['gr_HemeO'],  # heme-O
        'sheme_c': ['M00846'],   # siroheme
        'moco_c': ['M00880'],  # molybdenum cofactor
        'phllqol_c': ['M00932'],  # phylloquinol
        'gthrd_c': ['M00118'],  # Reduced glutathione
        
        ##### lipids #####
        #'pe120_c', # phosphatidyl-ethanolamine (12:0;12:0)
        #'pg120_c', # phosphatidyl-glycerol (12:0;12:0)
        #'clpn120_c', # cardiolipin (12:0;12:0;12:0;12:0)
        
        ##### membrane_wall #####
        # 1-lysyl phosphatidylglycerol (plantarum)
        'peptidoSTA_c': ['gr_ptdSTA'], # peptidoglycan (dependant on 'udcpdp_c')
        'WTAgg40r_20n_20a_P_c': ['gr_WTA3'], # teichoic acids
        'WTArg40r_20g_20a_P_c': ['gr_WTA2'], # teichoic acids
        'WTAg40g_20g_20a_P_c':  ['gr_WTA1'], # teichoic acids
        'LTAgg40g_20n_20a_c':  ['gr_LTA1'], # lipoteichoic acids
        'LTAga40g_20t_20a_c':  ['gr_LTA2'], # lipoteichoic acids            
        # capsular polysaccharides
        # kdo_lipid_A
        
    }
    modeled_rids = [r.id for r in model.reactions]
    
    
    cnt_removed = 0
    varprec = {}  # dictionary of variable biomass precursors
    for precursor, pathways in precursor_to_pathway.items(): 
        
        pathway_to_absence = {}
        pathway_to_compstring = {}   # completeness string
        for pathway in pathways:   # 2+ pathways might lead to the same precursor
            # initialize counters:
            cnt_members_tot = 0
            cnt_members_present = 0

            
            if pathway not in [gr.id for gr in universe.groups]:
                continue   # still missing from the universe
                
            for member in universe.groups.get_by_id(pathway).members:
                cnt_members_tot += 1    
                if member.id in modeled_rids:
                    cnt_members_present += 1
            # populate dicts:
            pathway_to_absence[pathway] = (cnt_members_present / cnt_members_tot) < 0.50
            pathway_to_compstring[pathway] = f'{pathway}: {cnt_members_present}/{cnt_members_tot}'
            

        varprec[precursor] = '; '.join(list(pathway_to_compstring.values()))
        if all(list(pathway_to_absence.values())):
            cnt_removed += 1
            logger.debug(f"Biomass precursor '{precursor}' seems not required ({varprec[precursor]}).")
            # add metabolites to the right side (they will disappear if the balance if 0)
            model.reactions.Biomass.add_metabolites({precursor: 0.01})

       
    logger.info(f"Removed {cnt_removed} biomass precursors.")
    return varprec




def get_gapfilling_repository(logger, universe, reference, exclude_orphans):
    
    
    # if provided, use the reference model as repository of reactions
    if reference != '-':   
        if reference.endswith('.xml'):
            refmodel = cobra.io.read_sbml_model(reference)
        elif reference.endswith('.json'):
            refmodel = cobra.io.load_json_model(reference)
        else:
            logger.error(f"Likely unsupported format found in --reference. Please use '.xml' or '.json'.")
            return 1
        repository = refmodel
    else:
        repository = universe


    # make an editable copy:
    repository_nogenes = repository.copy()


    if exclude_orphans:
        logger.info(f"Gapfilling is performed after removing orphan reactions.")
        to_remove = []

        cnt = 0
        for r in repository_nogenes.reactions: 
            if len(r.genes) !=0:
                continue 
            if len(r.metabolites) ==1:
                continue   # exclude exchanges/sinks/demands
            if any([m.id.endswith('_e') for m in r.metabolites]):
                continue   # exclude transporters

            cnt +=1
            to_remove.append(r)
            logger.debug(f"Removing orphan #{cnt}: {r.id} ({r.reaction}).")
        repository_nogenes.remove_reactions(to_remove)


    # remove genes to avoid the "ValueError: id purP is already present in list"
    cobra.manipulation.delete.remove_genes(repository_nogenes, [g.id for g in repository_nogenes.genes], remove_reactions=False)

    
    return repository_nogenes



def gapfill_on_media(logger, model, universe, expcon, media, varprec, reference, exclude_orphans):
    
    
    
    # get biomass precursors to gap-fill: 
    biomass_mids = []
    for value in get_biomass_dict().values(): biomass_mids = biomass_mids + value
    
    
    # prepare biomass sheet for excel output
    df_B = pnd.DataFrame()
    df_B['name'] = None
    df_B['cond'] = None
    df_B['coeff'] = None
    for mid in biomass_mids:
        
        df_B.loc[mid, 'name'] = universe.metabolites.get_by_id(mid).name
        df_B.loc[mid, 'coeff'] = 0.01
        if mid in varprec.keys():
            df_B.loc[mid, 'cond'] = varprec[mid]
            if mid not in [m.id for m in  model.reactions.Biomass.reactants]:  
                df_B.loc[mid, 'coeff'] = None
        else:  # biomass precursor is NOT variable (it's mandatory)
            df_B.loc[mid, 'cond'] = None
            
    
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: gap-filling will be skipped.")
        return df_B
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
        
        
    # get the repository of reactions:
    reference = '-'   # force universe: only universe will be used in this gap-filing step.
    repository_nogenes = get_gapfilling_repository(logger, universe, reference, exclude_orphans)
    
    
    logger.info(f"Gap-filling for biomass on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_B[f'{medium}'] = None
        logger.debug(f"Gap-filling on medium '{medium}'...")
                    
            
        
        # apply medium both on universe and model:
        response = apply_medium_given_column(logger, repository_nogenes, medium, expcon['media'][medium])
        if response == 1: return 1
        if not verify_growth(repository_nogenes):
            logger.error(f"Medium '{medium}' does not support growth of universe.")
            return 1
        
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1
        if verify_growth(model):
            logger.info(f"No need to gapfill model on medium '{medium}'.")
            continue


        # launch gap-filling separately for each biomass precursor:
        for mid in biomass_mids:
            if mid in [m.id for m in model.reactions.Biomass.reactants]:
            
                # save time if it can already be synthesized
                if gempipe.can_synth(model, mid)[0]:
                    df_B.loc[mid, f'{medium}'] = ''
                    logger.debug(f"Gap-filled 0 reactions on medium '{medium}' for '{mid}': [].")
                    continue   # save time!


                # otherwise perform the actual gap-fill:    
                suggested_rids = gempipe.perform_gapfilling(model, repository_nogenes, mid, nsol=1, verbose=False)
                df_B.loc[mid, f'{medium}'] = '; '.join(suggested_rids)
                logger.debug(f"Gap-filled {len(suggested_rids)} reactions on medium '{medium}' for '{mid}': {suggested_rids}.")
                for rid in suggested_rids:
                    gempipe.import_from_universe(model, repository_nogenes, rid, gpr='')
                
                
    return df_B

    
    
def focused_suggestions_on_media(logger, model, universe, expcon, media, suggest, reference, exclude_orphans):
    
    
    # get metabolites mids on which to focus: 
    if suggest == '-':
        return 0    
    focus_mids = suggest.split(',')
    modeld_mids = [m.id for m in model.metabolites]


    # get involved media:
    if media == '-':
        logger.info(f"No media provided: synthesis suggestions will be skipped.")
        return 0
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1


    # get the repository of reactions:
    repository_nogenes = get_gapfilling_repository(logger, universe, reference, exclude_orphans)


    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue

        
        # apply medium both on universe and model:
        if_reference = True if reference != '-' else False
        response = apply_medium_given_column(logger, repository_nogenes, medium, expcon['media'][medium], if_reference)
        if response == 1: return 1

        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1

            
        for mid in focus_mids:
            logger.info(f"Suggesting missing reactions for '{mid}' on  medium '{medium}'...")

            
            # focus on a particular metabolite:
            if not (mid in modeld_mids and mid.endswith('_c')):
                logger.info(f"Cytosolic metabolite defined with --suggest is not included: '{mid}'.")
                continue
            
            
            # repository model must be able to produce it:
            if not gempipe.can_synth(repository_nogenes, mid)[0]:
                logger.info(f"Medium '{medium}' does not support synthesis of '{mid}' in repository model.")
                continue

            
            nsol = 5   # number of solutions
            logger.info(f"Computing {nsol} solutions for cytosolic metabolite {mid}...")
            # model and universe are already set up with the same growth medium:
            print()
            # perform gap-filling, solutions are shown using print()
            _ = gempipe.perform_gapfilling(model, repository_nogenes, mid, nsol=nsol)
            print()

    
    return 0

    

    
def remove_disconnected(logger, model):
    
    to_remove = []
    for m in model.metabolites:
        if len(m.reactions) == 0:
            to_remove.append(m)
    model.remove_metabolites(to_remove)
    logger.info(f"Removed {len(to_remove)} disconnected metabolites.")
    
    

def write_excel_model(model, filepath, df_B, df_P):
    
    df_R = []
    df_M = []
    df_T = []
    df_A = []
    
    # format df_B:  # biomass assembly
    df_B.insert(0, 'mid', '')  # new columns as first
    df_B['mid'] = df_B.index
    df_B = df_B.reset_index(drop=True)
    
    # format df_P:  phenotype screening (Biolog(R))
    df_P.insert(0, 'plate:well', '')  # new columns as first
    df_P['plate:well'] = df_P.index
    df_P = df_P.reset_index(drop=True)
    
    
    
    for r in model.reactions:
        
        
        # handle artificial reactions
        if r.id == 'Biomass':
            df_A.append({'rid': r.id, 'rstring': r.reaction, 'type': 'biomass', 'name': r.name})
            
            
        elif len(r.metabolites) == 1:
            if len(r.metabolites)==1 and list(r.metabolites)[0].id.rsplit('_',1)[-1] == 'e': 
                df_A.append({'rid': r.id, 'rstring': r.reaction, 'type': 'exchange', 'name': r.name})
            elif r.lower_bound < 0 and r.upper_bound > 0:
                df_A.append({'rid': r.id, 'rstring': r.reaction, 'type': 'sink', 'name': r.name})
            elif r.lower_bound == 0 and r.upper_bound > 0:
                df_A.append({'rid': r.id, 'rstring': r.reaction, 'type': 'demand', 'name': r.name})
          
        
        else: # more than 1 metabolite involved
            
            # get kr codes: 
            if 'kegg.reaction' not in r.annotation.keys():  kr_ids = ''
            else:  
                kr_ids = r.annotation['kegg.reaction']
                if type(kr_ids) == str: kr_ids = [kr_ids]
                kr_ids = '; '.join([i for i in kr_ids if i!='RXXXXX'])

            # introduce reaction in the correct table: 
            r_dict = {'rid': r.id, 'rstring': r.reaction, 'kr': kr_ids, 'gpr': r.gene_reaction_rule, 'name': r.name}
            if len(set([m.id.rsplit('_',1)[-1] for m in r.metabolites])) == 1:
                df_R.append(r_dict)
            else: df_T.append(r_dict)
        
        
    for m in model.metabolites: 
        
        # get kc codes: 
        if 'kegg.compound' not in m.annotation.keys():  kc_ids = ''
        else:  
            kc_ids = m.annotation['kegg.compound']
            if type(kc_ids) == str: kc_ids = [kc_ids]
            kc_ids = '; '.join([i for i in kc_ids if i!='CXXXXX'])
        
        df_M.append({'mid': m.id, 'formula': m.formula, 'charge': m.charge, 'kc': kc_ids, 'name': m.name})

    
    
    df_R = pnd.DataFrame.from_records(df_R)
    df_M = pnd.DataFrame.from_records(df_M)
    df_T = pnd.DataFrame.from_records(df_T)
    df_A = pnd.DataFrame.from_records(df_A)
    with pnd.ExcelWriter(filepath) as writer:
        df_R.to_excel(writer, sheet_name='Reactions', index=False)
        df_M.to_excel(writer, sheet_name='Metabolites', index=False)
        df_T.to_excel(writer, sheet_name='Transporters', index=False)
        df_A.to_excel(writer, sheet_name='Artificials', index=False)
        df_B.to_excel(writer, sheet_name='Biomass', index=False)
        if len(df_P)!=0: df_P.to_excel(writer, sheet_name='Biolog®', index=False)



def biolog_on_media(logger, model, universe, expcon, media, biolog, reference, exclude_orphans):
    
    
    # prepare biomass sheet for excel output
    df_P = pnd.DataFrame()
    
    
    # load assets:
    official_pm_tables = {}
    for pm in ['PM1', 'PM2A', 'PM3B', 'PM4A']:
        with resources.path("gsrap.assets", f"{pm}.csv") as asset_path: 
            official_pm_tables[pm] = pnd.read_csv(asset_path, index_col=1, names=['plate','substrate','source','u1','u2','u3','u4','kc','cas'])
    
    
    # get involved media:
    if biolog == '-':
        return df_P
    if media == '-':
        logger.info(f"No media provided: Biolog(R)-based model curation will be skipped.")
        return df_P
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
    
    # get plates for this strain
    avail_plates = []
    for pm in ['PM1', 'PM2A', 'PM3B', 'PM4A']:
        if biolog in expcon[pm].columns: 
            avail_plates.append(pm)
    if avail_plates == set():
        logger.info(f"No Biolog(R) plates found for strain '{biolog}': Biolog(R)-based model curation will be skipped.")
        return df_P
    else:
        logger.info(f"Found {len(avail_plates)} Biolog(R) plates for strain '{biolog}': {sorted(avail_plates)}.")
        
        
    # get kc annotations:
    kc_to_exr = {}
    for m in model.metabolites: 
        if m.id.endswith('_e') == False:
            continue
        if 'kegg.compound' not in m.annotation.keys():
            continue
        kc_ids = m.annotation['kegg.compound']
        if type(kc_ids) == str: kc_ids = [kc_ids]
        kc_ids = [i for i in kc_ids if i!='CXXXXX']  
        for kc_id in kc_ids:
            kc_to_exr[kc_id] = f'EX_{m.id}'
        
        
    # prepare rows:
    for pm in avail_plates:
        for well, row in official_pm_tables[pm].iterrows():
            # write substrate name:
            df_P.loc[f"{pm}:{well}", 'substrate'] = row['substrate']
            
            
            # write source type:
            if pm in ['PM1', 'PM2A']: 
                df_P.loc[f"{pm}:{well}", 'source'] = 'carbon'
            elif pm == 'PM3B': 
                df_P.loc[f"{pm}:{well}", 'source'] = 'nitrogen'
            else:
                if well[0] in ['F','G','H']: df_P.loc[f"{pm}:{well}", 'source'] = 'sulfur'
                else: df_P.loc[f"{pm}:{well}", 'source'] = 'phosphorus'
            
            
            # get kc and write the correspondent exchange
            kc = row['kc']
            if type(kc)==float: 
                df_P.loc[f"{pm}:{well}", 'exchange'] = 'NA'
            elif kc.startswith('C'):
                if kc not in kc_to_exr.keys():
                    df_P.loc[f"{pm}:{well}", 'exchange'] = 'NA'
                else:
                    df_P.loc[f"{pm}:{well}", 'exchange'] = kc_to_exr[kc]
            elif kc.startswith('D'):
                df_P.loc[f"{pm}:{well}", 'exchange'] =  'NA'  # TODO manage exchanges in this case
            else:
                df_P.loc[f"{pm}:{well}", 'exchange'] =  '???'
                
                
    
    # get the repository of reactions:
    reference = '-'   # force universe: only universe will be used in this gap-filing step.
    repository_nogenes = get_gapfilling_repository(logger, universe, reference, exclude_orphans)
    
    
    logger.info(f"Performing Biolog(R)-based model curation on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_P[f'{medium}'] = None
        logger.debug(f"Performing Biolog(R)-based model curation on medium '{medium}'...")
        

    
    return df_P


import pandas as pnd
import cobra


import gempipe 

from ..commons import apply_medium_given_column
from ..commons import verify_growth

__OPTTHR__ = 0.001   # optimization threshold



def grow_on_media(logger, model, expcon, media, fva):
    obj_id = 'max(obj)'
    
    
    # prepare biomass sheet for excel output
    df_G = pnd.DataFrame()
    
    
    # get involved media:
    if media == '-':
        logger.info(f"No media provided: growth simulations will be skipped.")
        return df_G
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
    
    # initialize dataframe
    df_G.loc[obj_id, 'name'] = f'Objective reaction {gempipe.get_objectives(model)}'
    for r in model.reactions:
        df_G.loc[r.id, 'name'] = r.name
        
        
    logger.info(f"Growing on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column con the excel output
        df_G[f'{medium}'] = None
        logger.debug(f"Growing on medium '{medium}'...")
                    
        
        # apply medium
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1

    
        # perform FBA
        res_fba = verify_growth(model, boolean=False)
        df_G.loc[obj_id, f'{medium}'] = res_fba
        
        
        # perform FVA if requested:
        if fva and (res_fba not in [0, 'infeasible']):
            logger.debug(f"FVA on medium '{medium}'...")
            df_fva = cobra.flux_analysis.flux_variability_analysis(model, model.reactions, fraction_of_optimum=0.9)
            for rid, row in df_fva.iterrows():
                df_G.loc[rid, f'{medium}'] = f"({round(row['minimum'], 3)}, {round(row['maximum'], 3)})"

    
    return df_G

    

def synthesize_on_media(logger, model, expcon, media, synth):
    
    global __OPTTHR__
    
    # prepare biomass sheet for excel output
    df_S = pnd.DataFrame()
    
    
    # get involved media:
    if synth == False:
        return df_S
    if media == '-':
        logger.info(f"No media provided: biosynthesis test will be skipped.")
        return df_S
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
    
    # initialize dataframe
    logger.info(f"Testing synthesis on {len(media)} media...")
    for m in model.metabolites:
        df_S.loc[m.id, 'name'] = m.name
        
        
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column con the excel output
        df_S[f'{medium}'] = None
        logger.debug(f"Testing synthesis on medium '{medium}'...")
                    
        
        # apply medium
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1

    
        for m in model.metabolites:
            if m.id.endswith('_c'):
                binary, obj_value, status = gempipe.can_synth(model, m.id)
                if status =='infeasible':
                    df_S.loc[m.id, f'{medium}'] = 'infeasible'
                elif obj_value == 0:
                    df_S.loc[m.id, f'{medium}'] = 0
                elif obj_value < __OPTTHR__:
                    df_S.loc[m.id, f'{medium}'] = f'<{__OPTTHR__}'
                else:
                    df_S.loc[m.id, f'{medium}'] = round(obj_value, 3)

    
    return df_S



def omission_on_media(logger, model, expcon, media, omission):
        
    
    # define metabolites for single omission experiments  
    # Note: they are all universal
    pure_mids = set([
        # 20 aminoacids
        'ala__L', 'arg__L', 'asn__L', 'asp__L', 'cys__L', 
        'gln__L', 'glu__L', 'gly', 'his__L', 'ile__L', 
        'leu__L', 'lys__L', 'met__L', 'phe__L', 'pro__L', 
        'ser__L', 'thr__L', 'trp__L', 'tyr__L', 'val__L', 
        # 5 nucleotides
        'ade', 'gua', 'csn', 'thym', 'ura'])
    modeled_rids = [r.id for r in model.reactions]
    
    
    # prepare biomass sheet for excel output
    df_O = pnd.DataFrame()
    
    
    # get involved media:
    if omission == False:
        return df_O
    if media == '-':
        logger.info(f"No media provided: single omissions will be skipped.")
        return df_O
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
        
    # check if the target exchange reaction exists:
    for pure_mid in list(pure_mids):
        if f'EX_{pure_mid}_e' not in modeled_rids:
            logger.debug(f"Exchange reaction 'EX_{pure_mid}_e' not found during single omissions: it will be ignored.")
            df_O.loc[pure_mid, 'name'] = 'NA'
            pure_mids = pure_mids - set([pure_mid]) 
        else:  # initialize dataframe
            m = model.metabolites.get_by_id(f'{pure_mid}_e')
            df_O.loc[pure_mid, 'name'] = m.name
    
    
    logger.info(f"Performing single omissions on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_O[f'{medium}'] = None
        logger.debug(f"Performing single omissions on medium '{medium}'...")
        
        
        # apply medium
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1


        # set up exchange reactions:
        for pure_mid1 in pure_mids:
            for pure_mid2 in pure_mids:
                exr_id = f'EX_{pure_mid2}_e'
                
                if pure_mid2 != pure_mid1:
                    model.reactions.get_by_id(exr_id).lower_bound = -1000
                else:
                    model.reactions.get_by_id(exr_id).lower_bound = -0
                    
                    
            # perform FBA
            res_fba = verify_growth(model, boolean=False)
            df_O.loc[pure_mid1, f'{medium}'] = res_fba
               
            
    return df_O



def clean_formula(formula):
    
    # avoid confusion with 'C':
    formula = formula.replace('Ca', '').replace('Co', '').replace('Cu', '').replace('Cd', '').replace('Cr', '').replace('Cs', '').replace('Cl', '')   
    # avoid confusion with 'N':
    formula = formula.replace('Na', '').replace('Nb', '').replace('Ni', '').replace('Ne', '')
    # avoid confusion with 'P':
    formula = formula.replace('Pd', '').replace('Pt', '').replace('Pb', '').replace('Po', '')
    # avoid confusion with 'S':
    formula = formula.replace('Sc', '').replace('Si', '').replace('Sn', '').replace('Sb', '').replace('Se', '')
    
    return formula



def get_CNPS_sources(model):
    
    CNPS_sources = {'C': set(), 'N': set(), 'P': set(), 'S': set()}
    for r in model.reactions: 
        if len(r.metabolites)==1 and list(r.metabolites)[0].id.endswith('_e'):
            m = list(r.metabolites)[0]
            
            formula = m.formula
            formula = clean_formula(formula)
            
            if 'C' in formula: CNPS_sources['C'].add(r.id)
            if 'N' in formula: CNPS_sources['N'].add(r.id)
            if 'P' in formula: CNPS_sources['P'].add(r.id)
            if 'S' in formula: CNPS_sources['S'].add(r.id)
    
    return CNPS_sources


    
def get_source_name(atom):
    
    if atom=='C': return 'carbon'
    elif atom=='N': return 'nitrogen'
    elif atom=='P': return 'phosphorus'
    elif atom=='S': return 'sulfur'
    
    
    
def cnps_on_media(logger, model, expcon, media, cnps):
    
    
    # prepare biomass sheet for excel output
    df_C = pnd.DataFrame()
    
    
    # get involved media:
    if cnps == '-':
        return df_C
    if media == '-':
        logger.info(f"No media provided: alternative substrate analysis will be skipped.")
        return df_C
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
    
    # get main/starting C/N/P/S sources:
    if cnps == 'std': cnps = 'glc__D,nh4,pi,so4'
    startings = cnps.split(',')
    if len(startings) != 4: 
        logger.error(f"Starting sources must be 4: C, N, P, and S, in this order ({len(startings)} provided: {startings}).")
        return 1
    modeled_rids = [r.id for r in model.reactions]
    source_to_exr = {}
    for mid, source in zip(startings, ['C','N','P','S']):
        if mid.endswith('_e'): 
            mid_e = mid
        else: mid_e = mid + '_e'
        exr_id = f'EX_{mid_e}'
        if exr_id not in modeled_rids:
            logger.error(f"Expected exchange for {source} source not found (provided metabolite: '{mid}'; expected exchange: '{exr_id}').")
            return 1
        else:
            m = model.metabolites.get_by_id(mid_e)
            if source not in clean_formula(m.formula):
                logger.error(f"{source} source provided ('{mid}') does not contain {source} atoms.")
                return 1
            else:
                source_to_exr[source] = exr_id
            

    # define rows
    CNPS_sources = get_CNPS_sources(model)
    for source in ['C','N','P','S']:
        for exr_id in CNPS_sources[source]:
            m = list(model.reactions.get_by_id(exr_id).metabolites)[0]
            df_C.loc[f"[{source}] {exr_id}", 'source'] = get_source_name(source)
            df_C.loc[f"[{source}] {exr_id}", 'exchange'] = exr_id 
            df_C.loc[f"[{source}] {exr_id}", 'name'] = m.name    
    
    
    logger.info(f"Performing alternative substrate analysis on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_C[f'{medium}'] = None
        logger.debug(f"Performing alternative substrate analysis on medium '{medium}'...")
        
        
        not_part = set()
        for source in ['C','N','P','S']:
            for exr_id in CNPS_sources[source]:
                
                # apply medium
                response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
                if response == 1: return 1
            
                
                # maybe the starting substrate is not part of the medium
                if source_to_exr[source] not in list(model.medium.keys()):
                    not_part.add(f"[{source}] {source_to_exr[source]}")
                    df_C.loc[f"[{source}] {exr_id}", medium] = f"NA"
                    continue
                
            
                model.reactions.get_by_id(source_to_exr[source]).lower_bound = -0
                obj0 = verify_growth(model, boolean=False)
                model.reactions.get_by_id(exr_id).lower_bound = -1000
                obj1 = verify_growth(model, boolean=False)
                
                
                if obj1 == 'infeasible':
                    df_C.loc[f"[{source}] {exr_id}", medium] = 'infeasible'
                elif obj0 == 'infeasible':
                    df_C.loc[f"[{source}] {exr_id}", medium] = f"{obj1} (Δ=/)"
                else:
                    df_C.loc[f"[{source}] {exr_id}", medium] = f"{obj1} (Δ={obj1-obj0})"
        
    
        # log if some starting source was not contained in this medium
        if not_part != set(): 
            logger.debug(f"Specified starting sources {not_part} were not contained in medium '{medium}'.")
    
    return df_C
    


def essential_genes_on_media(logger, model, expcon, media, essential):
    
    global __OPTTHR__
    
    
    # prepare biomass sheet for excel output
    df_E = pnd.DataFrame()
    
    
    # get involved media:
    if essential == False:
        return df_E
    if media == '-':
        logger.info(f"No media provided: essential genes prediction will be skipped.")
        return df_E
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    

    # define rows
    for g in model.genes:
        if g.id in ['spontaneous', 'orphan']:
            continue
        df_E.loc[f"{g.id}", 'name'] = g.name    
        df_E.loc[f"{g.id}", 'involving'] = '; '.join([r.id for r in g.reactions])
    
    
    logger.info(f"Predicting essential genes on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_E[f'{medium}'] = None
        logger.debug(f"Predicting essential genes on medium '{medium}'...")
        
        
        # apply medium
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1
    
    
        # verify growth:
        obj = verify_growth(model, boolean=False)
        if obj in ['infeasible', 0]:
            df_E[f'{medium}'] = 'NA'
            continue
    
        
        single = cobra.flux_analysis.single_gene_deletion(model)
        for index, row in single.iterrows(): 
            gid = list(row['ids'])[0]  # they are single deletions
            if gid in ['spontaneous', 'orphan']:
                continue
                

            if row['status'] == 'infeasible': 
                df_E.loc[f"{gid}", medium] = 'True'
            elif row['growth'] < __OPTTHR__:
                df_E.loc[f"{gid}", medium] = 'True'
            else: 
                df_E.loc[f"{gid}", medium] = round(row['growth'], 3)
    
    return df_E



def growth_factors_on_media(logger, model, expcon, media, factors):
        
    
    # prepare biomass sheet for excel output
    df_F = pnd.DataFrame()
    
    
    # get involved media:
    if factors == False:
        return df_F
    if media == '-':
        logger.info(f"No media provided: essential genes prediction will be skipped.")
        return df_F
    media = media.split(',')
    # at least 1 medium must exist:
    if any([i in expcon['media'].columns for i in media])==False:
        logger.error(f"None of the provided media IDs exist. Available media are {list(expcon['media'].columns)}.")
        return 1
    
    
    # define rows
    df_F['mid'] = ''
    df_F['name'] = ''

    
    logger.info(f"Predicting growth factors on {len(media)} media...")
    for medium in media: 
        if medium not in expcon['media'].columns:
            logger.info(f"Medium '{medium}' does not exists and will be ignored.")
            continue
        # create dedicated column on the excel output
        df_F[f'{medium}'] = None
        logger.debug(f"Predicting growth factors on medium '{medium}'...")
        
        
        # apply medium
        response = apply_medium_given_column(logger, model, medium, expcon['media'][medium])
        if response == 1: return 1
    
    
        # verify growth:
        obj = verify_growth(model, boolean=False)
        if obj not in ['infeasible', 0]:
            df_F[f'{medium}'] = ''
        else:
            while obj in ['infeasible', 0]:
                res_dict = gempipe.sensitivity_analysis(model, scaled=False, top=1)
                for exr_id, value in res_dict.items():
                    if value < 0:
                        model.reactions.get_by_id(exr_id).lower_bound = -1000
                        obj = verify_growth(model, boolean=False)
                        df_F.loc[exr_id, f'{medium}'] = 'ADD'

         
    # populate 'mid'/'name' columns:
    for exr_id, row in df_F.iterrows():
        r = model.reactions.get_by_id(exr_id)
        m = list(r.metabolites)[0]
        df_F.loc[exr_id, 'mid'] = m.id.rsplit('_', 1)[0]
        df_F.loc[exr_id, 'name'] = m.name
    # replace index with 'mid':
    df_F = df_F.set_index('mid', drop=True)
        
        
    return df_F


    

def write_excel_model(model, filepath, df_G, df_S, df_O, df_C, df_E, df_F):
    

    
    # format df_G:  # growth tests
    df_G.insert(0, 'rid', '')  # new columns as first
    df_G['rid'] = df_G.index
    df_G = df_G.reset_index(drop=True)
    
    # format df_S:  # biosynthetic capabilities
    df_S.insert(0, 'mid', '')  # new columns as first
    df_S['mid'] = df_S.index
    df_S = df_S.reset_index(drop=True)
    
    # format df_O:  # single omission
    df_O.insert(0, 'mid', '')  # new columns as first
    df_O['mid'] = df_O.index
    df_O = df_O.reset_index(drop=True)
    
    # format df_C:  phenotype screening (CNPS)
    df_C.insert(0, 'substrate', '')  # new columns as first
    df_C['substrate'] = df_C.index
    df_C = df_C.reset_index(drop=True)
    
    # format df_E:  essential genes
    df_E.insert(0, 'gid', '')  # new columns as first
    df_E['gid'] = df_E.index
    df_E = df_E.reset_index(drop=True)
    
    # format df_F:  growth factors
    df_F.insert(0, 'mid', '')  # new columns as first
    df_F['mid'] = df_F.index
    df_F = df_F.reset_index(drop=True)
    
    
    
    with pnd.ExcelWriter(filepath) as writer:
        if len(df_G)!=0: df_G.to_excel(writer, sheet_name='Growth', index=False)
        if len(df_S)!=0: df_S.to_excel(writer, sheet_name='Synthesis', index=False)
        if len(df_O)!=0: df_O.to_excel(writer, sheet_name='Omission', index=False)
        if len(df_C)!=0: df_C.to_excel(writer, sheet_name='Substrates', index=False)
        if len(df_E)!=0: df_E.to_excel(writer, sheet_name='Essential', index=False)
        if len(df_F)!=0: df_F.to_excel(writer, sheet_name='Factors', index=False)





import os
from pathlib import Path


import gempipe
import cobra


from ..commons import get_expcon
from ..commons import force_id_on_sbml

from .unipruner import check_inputs
from .unipruner import parse_eggnog
from .unipruner import subtract_kos
from .unipruner import translate_remaining_kos
from .unipruner import restore_gene_annotations
from .unipruner import include_forced
from .unipruner import adjust_biomass_equation
from .unipruner import gapfill_on_media
from .unipruner import biolog_on_media
from .unipruner import remove_disconnected
from .unipruner import write_excel_model
from .unipruner import log_metrics
from .unipruner import focused_suggestions_on_media



def main(args, logger):
    
    
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    
    
    
    # check input files:
    response = check_inputs(logger, args.universe, args.eggnog)
    if type(response)==int:
        return 1
    universe = response[0]
    eggnog = response[1]
        
    
    # check file structure ('expcon'= experimental constraints)
    expcon = get_expcon(logger)
    if type(expcon)==int: return 1

    
    
    ###### PRUNING
    # get important dictionaries: 'eggnog_ko_to_gids' and 'eggonog_gid_to_kos'
    eggnog_ko_to_gids, eggonog_gid_to_kos = parse_eggnog(eggnog)
    
    # create a copy of the universe
    model = universe.copy()
    model.id = Path(args.eggnog).stem        
    
    # prune reactions
    subtract_kos(logger, model, eggnog_ko_to_gids)
    
    # translate KOs to the actual genes
    translate_remaining_kos(logger, model, eggnog_ko_to_gids)
    restore_gene_annotations(logger, model, universe, eggonog_gid_to_kos)
    
    
    
    ###### GAPFILLING
    # force inclusion of reactions:  
    include_forced(logger, model, universe, args.force_inclusion)
    #
    # adjust biomass equation:
    # get the variable precursors dataframe ('variable_precs_df')
    varprec = adjust_biomass_equation(logger, model, universe)
    #
    # gap-fill based on media:
    df_B = gapfill_on_media(logger, model, universe, expcon, args.gap_fill, varprec, args.reference, args.exclude_orphans)
    if type(df_B)==int: return 1
    #
    # perform Biolog(R) simulations based on media. Gap-filling is included. 
    df_P = biolog_on_media(logger, model, universe, expcon, args.gap_fill, args.biolog, args.reference, args.exclude_orphans)
    if type(df_P)==int: return 1
    

    
    # remove disconnected metabolites
    remove_disconnected(logger, model)
    
    
    # reset growth environment befor saving the model
    gempipe.reset_growth_env(model)
                     
        
    # output the model:
    cobra.io.save_json_model(model, f'{args.outdir}/{model.id}.json')        # JSON
    cobra.io.write_sbml_model(model, f'{args.outdir}/{model.id}.xml')        # SBML   # groups are saved only to SBML
    force_id_on_sbml(f'{args.outdir}/{model.id}.xml', model.id)   # force introduction of the 'id=""' field
    write_excel_model(model, f'{args.outdir}/{model.id}.mkmodel.xlsx', df_B, df_P)  
    logger.info(f"'{args.outdir}/{model.id}.json' created!")
    logger.info(f"'{args.outdir}/{model.id}.xml' created!")
    logger.info(f"'{args.outdir}/{model.id}.mkmodel.xlsx' created!")
    log_metrics(logger, model, outmode='recon_model')
    
    
    # suggest missing reactions focusing of provided metabolites
    response = focused_suggestions_on_media(logger, model, universe, expcon, args.gap_fill, args.suggest, args.reference, args.exclude_orphans)
    if response==1: return 1
    
    
    return 0


import os
import warnings
import logging


import pandas as pnd
import cobra


import gempipe

from ..commons import get_expcon
from ..commons import log_metrics

from .simarena import grow_on_media
from .simarena import synthesize_on_media
from .simarena import omission_on_media
from .simarena import cnps_on_media
from .simarena import essential_genes_on_media
from .simarena import growth_factors_on_media
from .simarena import write_excel_model



def main(args, logger):

    
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    
    
    # check file structure ('expcon'= experimental constraints)
    expcon = get_expcon(logger)
    if type(expcon)==int: return 1


    # load model
    logger.info("Reading input model...")
    model = None
    if args.inmodel.endswith('.xml'):
        # avoids warnings like "'Mendoza.WCFS1.emapper' is not a valid SBML 'SId'.":
        # As ~90 loggers are active, not knowing the actual one, temporarly silence every logger: 
        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        previous_levels = {logger: logger.level for logger in loggers}
        for logger in loggers:
            logger.setLevel(logging.CRITICAL)  # setting to "logging.CRITICAL" is not sufficient
        model = cobra.io.read_sbml_model(args.inmodel)  # read SBML with invalid 'SId'
        for logger, level in previous_levels.items():
            logger.setLevel(level)
    elif args.inmodel.endswith('.json'):
        model = cobra.io.load_json_model(args.inmodel)
    log_metrics(logger, model, outmode='loaded_model')
    
    
    # reset growth environment befor saving the model
    gempipe.reset_growth_env(model)
        
        
    # check presence of input media
    if args.fba == '-':
        logger.error("No media provided! Cannot do growth simulations without growth media.")
        return 1
        
        
    # growth simulations based on media
    df_G = grow_on_media(logger, model, expcon, args.fba, args.fva)
    if type(df_G)==int: return 1
    
    # test synthesis based on media
    df_S = synthesize_on_media(logger, model, expcon, args.fba, args.synth)
    if type(df_S)==int: return 1
    
    # perform single omission experiments based on media
    df_O = omission_on_media(logger, model, expcon, args.fba, args.omission)
    if type(df_O)==int: return 1
    
    # perform alternative substrate anaysis based on media
    df_C = cnps_on_media(logger, model, expcon, args.fba, args.cnps)
    if type(df_C)==int: return 1
    
    # predict essential genes
    df_E = essential_genes_on_media(logger, model, expcon, args.fba, args.essential)
    if type(df_E)==int: return 1
    
    # predict growth factors
    df_F = growth_factors_on_media(logger, model, expcon, args.fba, args.factors)
    if type(df_F)==int: return 1
        
    
    # save simulations:
    write_excel_model(model, f'{args.outdir}/{model.id}.runsims.xlsx',  df_G, df_S, df_O, df_C, df_E, df_F)  
    logger.info(f"'{args.outdir}/{model.id}.runsims.xlsx' created!")

    
    return 0


import os
import pickle


import cobra


from ..commons import log_metrics

from .tsiparser import get_db
from .tsiparser import introduce_metabolites
from .tsiparser import introduce_reactions
from .tsiparser import introduce_transporters
from .tsiparser import introduce_sinks_demands
from .tsiparser import introduce_biomass
from .tsiparser import translate_annotate_genes
from .tsiparser import set_up_groups
from .tsiparser import check_biosynthesis
from .tsiparser import check_completeness
from .tsiparser import show_contributions



def main(args, logger):
    
    
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    
    
    # check and extract the required 'gsrap.maps' file
    if os.path.exists(f'{args.inmaps}') == False:
        logger.error(f"File 'gsrap.maps' not found at {args.inmaps}.")
        return 1
    try: 
        with open(f'{args.inmaps}', 'rb') as f:
            inmaps = pickle.load(f)  
    except: 
        logger.error(f"Provided file {args.inmaps} has an incorrect format.")
        return 1
    idcollection_dict = inmaps['idcollection_dict']
    summary_dict = inmaps['summary_dict']
    
    
    # check compatibility of input parameters
    if args.progress==False and args.module==True: 
        logger.error(f"You cannot ask --module without --progress (see --help).")
        return 1
    if args.progress==False and args.focus!='-':
        logger.error(f"You cannot ask --focus without --progress (see --help).")
        return 1
    if args.progress==False and args.zeroes==True:
        logger.error(f"You cannot ask --zeroes without --progress (see --help).")
        return 1
    
    
    # check goodbefore and onlyauthor
    if args.goodbefore == '-' and args.onlyauthor != '-':
        logger.error(f"--onlyauthor must be used in conjunction with --goodbefore.")
        return 1
    if   args.goodbefore == '-': args.goodbefore = [None, None, None]
    elif len(args.goodbefore.split('-')) != 3: 
        logger.error(f"Invalid syntax detected for --goodbefore.")
        return 1
    else:
        args.goodbefore = args.goodbefore.split('-')
        if args.goodbefore[0] == 'None': args.goodbefore[0] = None
        if args.goodbefore[1] == 'None': args.goodbefore[1] = None
        if args.goodbefore[2] == 'None': args.goodbefore[2] = None
    if args.onlyauthor == '-': args.onlyauthor = None
    
    
    # download database and check its structure
    db = get_db(logger)
    if type(db)==int: return 1
                                    
        
    # create the model
    model = cobra.Model('newuni')
        
    
    # introduce M / R / T
    model = introduce_metabolites(logger, db, model, idcollection_dict, args.goodbefore[0], args.onlyauthor)
    if type(model)==int: return 1
    model = introduce_reactions(logger, db, model, idcollection_dict, args.goodbefore[1], args.onlyauthor)
    if type(model)==int: return 1
    model = introduce_transporters(logger, db, model, idcollection_dict, args.goodbefore[2], args.onlyauthor)
    if type(model)==int: return 1


    # introduce sinks / demands (exchanges where included during T)
    model = introduce_sinks_demands(logger, model)
    if type(model)==int: return 1


    # introducce biomass
    model = introduce_biomass(logger, db, model)
    if type(model)==int: return 1


    # translate Gs to symbols and annotate them (EC, COG, GO, ...)
    model = translate_annotate_genes(logger, model, idcollection_dict)
    if type(model)==int: return 1


    # introduce collectionas (groups of Rs as maps/modules)
    model = set_up_groups(logger, model, idcollection_dict)
    if type(model)==int: return 1
    
    
    # output the universe
    cobra.io.save_json_model(model, 'newuni.json')
    cobra.io.write_sbml_model(model, 'newuni.xml')   # groups are saved only to SBML 
    logger.info(f"'{args.outdir}/newuni.json' created!")
    logger.info(f"'{args.outdir}/newuni.xml' created!")
    log_metrics(logger, model, outmode='uni_features')

    
    response = check_biosynthesis(logger, model, args.outdir, args.growth, args.biosynth)
    if response==1: return 1
    
    
    response = check_completeness(logger, model, args.progress, args.module, args.focus, args.eggnog, args.zeroes, idcollection_dict, summary_dict)
    if response==1: return 1


    # show simple statistics of contributions
    show_contributions(logger, db)

    
      
    return 0


import shutil
import os
import pickle


from .kdown import download_raw_txtfiles
from .kdown import create_dict_ko
from .kdown import create_dict_c
from .kdown import create_dict_r
from .kdown import create_dict_map
from .kdown import create_dict_md
from .kdown import create_idcollection_dict
from .kdown import create_summary_dict



def main(args, logger):
    
    
    # adjust out folder path                      
    while args.outdir.endswith('/'):              
        args.outdir = args.outdir[:-1]
    os.makedirs(f'{args.outdir}/', exist_ok=True)
    

        
    logger.info(f"Respectfully retrieving metabolic information from KEGG. Raw data are being saved into '{args.outdir}/kdown/'. Be patient, could take a couple of days...")
    os.makedirs(f'{args.outdir}/kdown/', exist_ok=True)
    
    
    response = download_raw_txtfiles(logger, args.outdir , args.usecache)
    if type(response) == int: return 1
    else: RELEASE_kegg = response

    response = create_dict_ko(logger, args.outdir )
    if type(response) == int: return 1
    else: dict_ko = response
    
    response = create_dict_c(logger, args.outdir )
    if type(response) == int: return 1
    else: dict_c = response
    
    response = create_dict_r(logger, args.outdir )
    if type(response) == int: return 1
    else: dict_r = response
    
    response = create_dict_map(logger, args.outdir )
    if type(response) == int: return 1
    else: dict_map = response
    
    response = create_dict_md(logger, args.outdir )
    if type(response) == int: return 1
    else: dict_md = response
    
    
    # create 'gsrap.maps':
    idcollection_dict = create_idcollection_dict(dict_ko, dict_c, dict_r, dict_map, dict_md)
    summary_dict = create_summary_dict(dict_c, dict_r, dict_map, dict_md)
    with open(f'{args.outdir}/gsrap.maps', 'wb') as wb_handler:
        pickle.dump({'RELEASE_kegg': RELEASE_kegg, 'idcollection_dict': idcollection_dict, 'summary_dict': summary_dict}, wb_handler)
    logger.info(f"'{args.outdir}/gsrap.maps' created!")
    
        
    # clean temporary files:
    if not args.keeptmp:
        shutil.rmtree(f'{args.outdir}/kdown', ignore_errors=True)
        logger.info(f"Temporary raw files deleted!")

    
    return 0


import argparse
import sys
import multiprocessing 
import logging 
from logging.handlers import QueueHandler
import traceback
import importlib.metadata
from datetime import datetime


from .getmaps import getmaps_command
from .parsedb import parsedb_command
from .mkmodel import mkmodel_command
from .runsims import runsims_command


import cobra


cobra_config = cobra.Configuration()
solver_name = str(cobra_config.solver.log).split(' ')[1]
solver_name = solver_name.replace("optlang.", '')
solver_name = solver_name.replace("_interface", '')


def main():
    
    
    # define the header of main- and sub-commands. 
    header = f'gsrap v{importlib.metadata.metadata("gsrap")["Version"]},\ndeveloped by Gioele Lazzari (gioele.lazzari@univr.it).'
    
    
    # create the command line arguments:
    parser = argparse.ArgumentParser(description=header, add_help=False)
    parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    subparsers = parser.add_subparsers(title='gsrap subcommands', dest='subcommand', help='', required=True)
    
    
    # create the 3 subparsers:
    getmaps_parser = subparsers.add_parser('getmaps', description=header, help="Create 'gsrap.maps' by retrieving metabolic information from metabolic databases.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    parsedb_parser = subparsers.add_parser('parsedb', description=header, help="Parse the online database to create the universe GSMM. Requires 'gsrap.maps'.", formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    mkmodel_parser = subparsers.add_parser('mkmodel', description=header, help='Reconstruct a GSMM from the universe and a functional annotation.', formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)
    runsims_parser = subparsers.add_parser('runsims', description=header, help='Analyze a GSMM previously built.', formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)

    
    # add arguments for the 'getmaps' command
    getmaps_parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    getmaps_parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    getmaps_parser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
    getmaps_parser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
    getmaps_parser.add_argument("--keeptmp", action='store_true', help="Do not delete temporary raw files.")
    getmaps_parser.add_argument("--usecache", action='store_true', help="Do not update previously downloaded records (save time).")
    
    
    # add arguments for the 'parsedb' command
    parsedb_parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    parsedb_parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    parsedb_parser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
    parsedb_parser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
    parsedb_parser.add_argument("-i", "--inmaps", metavar='', type=str, default='./gsrap.maps', help="Input file 'gsrap.maps' previously produced using the 'getmaps' subcommand.")
    parsedb_parser.add_argument("-p", "--progress", action='store_true', help="Show progress for each map.")
    parsedb_parser.add_argument("-m", "--module", action='store_true', help="Show progress for each module of each map (use only with --progress).")
    parsedb_parser.add_argument("-f", "--focus", metavar='', type=str, default='-', help="Focus on a particular map/module (use only with --progress).")
    parsedb_parser.add_argument("-g", "--growth", action='store_true', help="Test growth on a minimal medium.")
    parsedb_parser.add_argument("-b", "--biosynth", metavar='', type=str, default='-',  help="Test biosynthesis of metabolites.")
    parsedb_parser.add_argument("-e", "--eggnog", metavar='', type=str, default='-', help="Path to the optional eggnog-mapper annotation table.")
    parsedb_parser.add_argument("-z", "--zeroes", action='store_true', help="Show maps/modules with 0%% coverage, in addition to partials (use only with --progress).")
    parsedb_parser.add_argument("-u", "--goodbefore", metavar='', type=str, default='-', help="Syntax is {pure_mid}-{rid1}-{rid2}. From top to bottom, build the universe until reaction {rid1}, transport {rid2} and metabolite {pure_mid} are reached.")
    parsedb_parser.add_argument("-a", "--onlyauthor", metavar='', type=str, default='-', help="Build the universe by parsing contents of the specified author ID only. Contents affected by --goodbefore are parsed anyway.")
    
    
    # add arguments for the 'mkmodel' command
    mkmodel_parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    mkmodel_parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    mkmodel_parser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
    mkmodel_parser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
    mkmodel_parser.add_argument("-e", "--eggnog", metavar='', type=str, default='-', help="Path to the eggnog-mapper annotation table.")
    mkmodel_parser.add_argument("-u", "--universe", metavar='', type=str, default='-', help="Path to the universe model (SBML format).")
    mkmodel_parser.add_argument("-i", "--force_inclusion", metavar='', type=str, default='-', help="Force the inclusion of the specified reactions (comma-separated IDs).")
    mkmodel_parser.add_argument("-f", "--gap_fill", metavar='', type=str, default='-', help="Media to use during gap-filling (comma-separated IDs); if not provided, gap-filling will be skipped.")
    mkmodel_parser.add_argument("-r", "--reference", metavar='', type=str, default='-', help="Use the specified model for gap-filling, instead of the universe.")
    mkmodel_parser.add_argument("-x", "--exclude_orphans", action='store_true', help="Exclude orphan reactions from the gap-filling repository.")
    mkmodel_parser.add_argument("-b", "--biolog", metavar='', type=str, default='-', help="Strain ID associated to binary Biolog(R) PM1, PM2A, PM3B and PM4A plates; if not provided, Biolog(R)-based model curation will be skipped (use with --cnps and --gap_fill).")
    mkmodel_parser.add_argument("--suggest", metavar='', type=str, default='-', help="Suggest missing reactions to synthetize the specified metabolites (comma-separated IDs) (use with --growth).")
    

    # add arguments for the 'runsims' command
    runsims_parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    runsims_parser.add_argument("-v", "--version", action="version", version=f"v{importlib.metadata.metadata('gsrap')['Version']}", help="Show version number and exit.")
    runsims_parser.add_argument("--verbose", action='store_true', help="Make stdout messages more verbose, including debug messages.")
    runsims_parser.add_argument("-o", "--outdir", metavar='', type=str, default='./', help="Main output directory (will be created if not existing).")
    runsims_parser.add_argument("-i", "--inmodel", metavar='', type=str, default='-', help="Input strain-specific GSMM.")
    runsims_parser.add_argument("-g", "--fba", "--growth", metavar='', type=str, default='-', help="Media to use during growth simulations (comma-separated IDs); if not provided, growth simulations will be skipped.")
    runsims_parser.add_argument("--fva", action='store_true', help="Perform FVA during growth simulations (use with --growth).")
    runsims_parser.add_argument("--synth", action='store_true', help="Test metabolites' synthesis on media (use with --growth).")
    runsims_parser.add_argument("-s", "--cnps", metavar='', type=str, default='-', help="Starting C, N, P and S source metabolites (comma-separated IDs); if not provided, anternative substrate analysis will be skipped.")
    runsims_parser.add_argument("-b", "--biolog", metavar='', type=str, default='-', help="Strain ID associated to binary Biolog(R) PM1, PM2A, PM3B and PM4A plates; if not provided, Biolog(R)-based model curation will be skipped (use with --cnps and --gap_fill).")
    runsims_parser.add_argument("--omission", action='store_true', help="Perform single omission experiments to study auxotrophies (use with --growth).")
    runsims_parser.add_argument("--essential", action='store_true', help="Predict essential genes (single-gene knock-out simulations) (use with --growth).")
    runsims_parser.add_argument("--factors", action='store_true', help="Predict putative growth factors (use with --growth).")

    

    # check the inputted subcommand, automatic sys.exit(1) if a bad subprogram was specied. 
    args = parser.parse_args()
    
    
    # set the multiprocessing context
    multiprocessing.set_start_method('fork') 
    
    
    # create a logging queue in a dedicated process.
    def logger_process_target(queue):
        logger = logging.getLogger('gsrap')
        while True:
            message = queue.get() # block until a new message arrives
            if message is None: # sentinel message to exit the loop
                break
            logger.handle(message)
    queue = multiprocessing.Queue()
    logger_process = multiprocessing.Process(target=logger_process_target, args=(queue,))
    logger_process.start()
    
    
    # connect the logger for this (main) process: 
    logger = logging.getLogger('gsrap')
    logger.addHandler(QueueHandler(queue))
    if args.verbose: logger.setLevel(logging.DEBUG) # debug (lvl 10) and up
    else: logger.setLevel(logging.INFO) # debug (lvl 20) and up
    
    
    # handy function to print without time/level (for header / trailer)
    def set_header_trailer_formatter(logger):
        formatter = logging.Formatter('%(message)s')
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return handler
    
    
    # to print the main pipeline logging:
    def set_usual_formatter(logger):
        formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s', datefmt="%H:%M:%S")
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return handler
    
    
    
    # show a welcome message:
    thf_handler = set_header_trailer_formatter(logger)
    logger.info(header + '\n')
    command_line = '' # print the full command line:
    for arg, value in vars(args).items():
        if arg == 'subcommand': command_line = command_line + f"gsrap {value} "
        else: command_line = command_line + f"--{arg} {value} "
    logger.info('Inputted command line: "' + command_line.rstrip() + '".\n')
    logger.removeHandler(thf_handler)
    
    
    
    usual_handler = set_usual_formatter(logger)
    current_date_time = datetime.now()
    formatted_date = current_date_time.strftime("%Y-%m-%d")
    logger.info(f"Welcome to 'gsrap {args.subcommand}'! Launching the tool on {formatted_date}...")
    logger.debug(f'COBRApy started with solver: {solver_name}.')
    try: 
        # choose which subcommand to lauch: 
        if args.subcommand == 'getmaps':
            response = getmaps_command(args, logger)
        if args.subcommand == 'parsedb':
            response = parsedb_command(args, logger)
        if args.subcommand == 'mkmodel':
            response = mkmodel_command(args, logger)
        if args.subcommand == 'runsims':
            response = runsims_command(args, logger)
            
        if response == 0:
            logger.info(f"'gsrap {args.subcommand}' terminated without errors!")
    except: 
        # show the error stack trace for this un-handled error: 
        response = 1
        logger.error(traceback.format_exc())
    logger.removeHandler(usual_handler)


    
    # Terminate the program:
    thf_handler = set_header_trailer_formatter(logger)
    if response == 1: 
        queue.put(None) # send the sentinel message
        logger_process.join() # wait for all logs to be digested
        sys.exit(1)
    else: 
        # show a bye message
        queue.put(None) # send the sentinel message
        logger_process.join() # wait for all logs to be digested
        logger.info('\n' + header)
        sys.exit(0) # exit without errors
        
        
        
if __name__ == "__main__":
    main()